from __future__ import absolute_import
from __future__ import print_function
import os
import sys
import collections
import functools

import veriloggen.core.vtypes as vtypes
from veriloggen.core.module import Module
from veriloggen.core.submodule import Submodule
from veriloggen.seq.subst_visitor import *
from veriloggen.seq.reset_visitor import ResetVisitor
from veriloggen.seq.seq import Seq, make_condition


_tmp_count = 0


def reset():
    global _tmp_count
    _tmp_count = 0


def _tmp_name(prefix='_tmp_fsm'):
    global _tmp_count
    v = _tmp_count
    _tmp_count += 1
    ret = '_'.join([prefix, str(v)])
    return ret


def TmpFSM(m, clk, rst, width=32, initname='init',
           nohook=False, as_module=False, prefix=None):
    if prefix is None:
        prefix = '_tmp_fsm'
    name = _tmp_name(prefix)
    return FSM(m, name, clk, rst, width, initname, nohook, as_module)


class FSM(vtypes.VeriloggenNode):
    """ Finite State Machine Generator """

    def __init__(self, m, name, clk, rst, width=32, initname='init',
                 nohook=False, as_module=False):
        self.m = m
        self.name = name
        self.clk = clk
        self.rst = rst
        self.width = width
        self.state_count = 0
        self.state = self.m.Reg(name, width)  # set initval later

        self.mark = collections.OrderedDict()  # key:index
        self._set_mark(0, self.name + '_' + initname)
        self.state.initval = self._get_mark(0)

        self.body = collections.defaultdict(list)
        self.jump = collections.defaultdict(list)

        self.delay_amount = 0
        self.delayed_state = collections.OrderedDict()  # key:delay
        self.delayed_body = collections.defaultdict(
            functools.partial(collections.defaultdict, list))  # key:delay
        self.delayed_cond = collections.OrderedDict()  # key:name
        self.tmp_count = 0

        self.dst_var = collections.OrderedDict()
        self.dst_visitor = SubstDstVisitor()
        self.reset_visitor = ResetVisitor()

        self.seq = Seq(self.m, self.name + '_par', clk, rst, nohook=True)

        self.done = False

        self.last_cond = []
        self.last_kwargs = {}
        self.last_if_statement = None
        self.elif_cond = None
        self.next_kwargs = {}

        self.as_module = as_module

        if not nohook:
            self.m.add_hook(self.implement)

    # -------------------------------------------------------------------------
    def goto(self, dst, cond=None, else_dst=None):
        if cond is None and 'cond' in self.next_kwargs:
            cond = self.next_kwargs['cond']
        self._clear_next_kwargs()
        self._clear_last_if_statement()
        self._clear_last_cond()

        src = self.current
        return self._go(src, dst, cond, else_dst)

    def goto_init(self, cond=None):
        if cond is None and 'cond' in self.next_kwargs:
            cond = self.next_kwargs['cond']
        self._clear_next_kwargs()
        self._clear_last_if_statement()
        self._clear_last_cond()

        src = self.current
        dst = 0
        return self._go(src, dst, cond)

    def goto_next(self, cond=None):
        if cond is None and 'cond' in self.next_kwargs:
            cond = self.next_kwargs['cond']
        self._clear_next_kwargs()
        self._clear_last_if_statement()
        self._clear_last_cond()

        src = self.current
        dst = self.current + 1
        ret = self._go(src, dst, cond=cond)
        self.inc()
        return ret

    def goto_from(self, src, dst, cond=None, else_dst=None):
        if cond is None and 'cond' in self.next_kwargs:
            cond = self.next_kwargs['cond']
        self._clear_next_kwargs()
        self._clear_last_if_statement()
        self._clear_last_cond()

        return self._go(src, dst, cond, else_dst)

    def inc(self):
        self._set_index(None)

    # -------------------------------------------------------------------------
    def add(self, *statement, **kwargs):
        """ add new assignments """
        kwargs.update(self.next_kwargs)
        self.last_kwargs = kwargs
        self._clear_next_kwargs()

        # if there is no attributes, Elif object is reused.
        has_args = not (len(kwargs) == 0 or  # has no args
                        (len(kwargs) == 1 and 'cond' in kwargs))  # has only 'cond'

        if self.elif_cond is not None and not has_args:
            next_call = self.last_if_statement.Elif(self.elif_cond)
            next_call(*statement)
            self.last_if_statement = next_call
            self._add_dst_var(statement)
            self._clear_elif_cond()
            return self

        self._clear_last_if_statement()
        return self._add_statement(statement, **kwargs)

    # -------------------------------------------------------------------------
    def add_reset(self, v):
        return self.seq.add_reset(v)

    # -------------------------------------------------------------------------
    def Prev(self, var, delay, initval=0, cond=None, prefix=None):
        return self.seq.Prev(var, delay, initval, cond, prefix)

    # -------------------------------------------------------------------------
    def If(self, *cond):
        self._clear_elif_cond()

        cond = make_condition(*cond)

        if cond is None:
            return self

        if 'cond' not in self.next_kwargs:
            self.next_kwargs['cond'] = cond
        else:
            self.next_kwargs['cond'] = vtypes.Ands(
                self.next_kwargs['cond'], cond)

        self.last_cond = [self.next_kwargs['cond']]

        return self

    def Else(self, *statement, **kwargs):
        self._clear_elif_cond()

        if len(self.last_cond) == 0:
            raise ValueError("No previous condition for Else.")

        old = self.last_cond.pop()
        self.last_cond.append(vtypes.Not(old))

        # if the true-statement has delay attributes,
        # Else statement is separated.
        if 'delay' in self.last_kwargs and self.last_kwargs['delay'] > 0:
            prev_cond = self.last_cond
            ret = self.Then()(*statement)
            self.last_cond = prev_cond
            return ret

        # if there is additional attribute, Else statement is separated.
        has_args = not (len(self.next_kwargs) == 0 or  # has no args
                        (len(self.next_kwargs) == 1 and 'cond' in kwargs))  # has only 'cond'

        if has_args:
            prev_cond = self.last_cond
            ret = self.Then()(*statement)
            self.last_cond = prev_cond
            return ret

        if not isinstance(self.last_if_statement, vtypes.If):
            raise ValueError("Last if-statement is not If")

        self.last_if_statement.Else(*statement)
        self._add_dst_var(statement)

        return self

    def Elif(self, *cond):
        if len(self.last_cond) == 0:
            raise ValueError("No previous condition for Else.")

        cond = make_condition(*cond)

        old = self.last_cond.pop()
        self.last_cond.append(vtypes.Not(old))
        self.last_cond.append(cond)

        # if the true-statement has delay attributes, Else statement is
        # separated.
        if 'delay' in self.last_kwargs and self.last_kwargs['delay'] > 0:
            prev_cond = self.last_cond
            ret = self.Then()
            self.last_cond = prev_cond
            return ret

        if not isinstance(self.last_if_statement, vtypes.If):
            raise ValueError("Last if-statement is not If")

        self.elif_cond = cond

        cond = self._make_cond(self.last_cond)
        self.next_kwargs['cond'] = cond

        return self

    def Delay(self, delay):
        self.next_kwargs['delay'] = delay
        return self

    def Keep(self, keep):
        self.next_kwargs['keep'] = keep
        return self

    def Then(self):
        cond = self._make_cond(self.last_cond)
        self._clear_last_cond()
        self.If(cond)
        return self

    def LazyCond(self, value=True):
        self.next_kwargs['lazy_cond'] = value
        return self

    def EagerVal(self, value=True):
        self.next_kwargs['eager_val'] = value
        return self

    def Clear(self):
        self._clear_next_kwargs()
        self._clear_last_if_statement()
        self._clear_last_cond()
        self._clear_elif_cond()
        return self

    # -------------------------------------------------------------------------
    @property
    def current(self):
        return self.state_count

    @property
    def next(self):
        return self.current + 1

    @property
    def current_delay(self):
        if 'delay' in self.next_kwargs:
            return self.next_kwargs['delay']
        return 0

    @property
    def last_delay(self):
        if 'delay' in self.last_kwargs:
            return self.last_kwargs['delay']
        return 0

    @property
    def current_condition(self):
        cond = self.next_kwargs['cond'] if 'cond' in self.next_kwargs else None
        if cond is not None:
            cond = vtypes.AndList(self.state == self.state_count, cond)
        else:
            cond = self.state == self.state_count
        return cond

    @property
    def last_condition(self):
        cond = self._make_cond(self.last_cond)
        if cond is not None:
            cond = vtypes.AndList(self.state == self.state_count, cond)
        else:
            cond = self.state == self.state_count
        return cond

    @property
    def then(self):
        return self.last_condition

    @property
    def here(self):
        return self.state == self.current

    # -------------------------------------------------------------------------
    def implement(self):
        if self.as_module:
            self.make_module()
            return

        self.make_always()

    # -------------------------------------------------------------------------
    def make_always(self, reset=(), body=(), case=True):
        if self.done:
            #raise ValueError('make_always() has been already called.')
            return

        self.done = True

        part_reset = self.make_reset(reset)
        part_body = list(body) + list(self.make_case()
                                      if case else self.make_if())
        self.m.Always(vtypes.Posedge(self.clk))(
            vtypes.If(self.rst)(
                part_reset,
            )(
                part_body,
            ))

    # -------------------------------------------------------------------------
    def make_module(self, reset=(), body=(), case=True):
        if self.done:
            #raise ValueError('make_always() has been already called.')
            return

        self.done = True

        m = Module('sub_%s' % self.name)

        clk = m.Input('CLK')

        if self.rst is not None:
            rst = m.Input('RST')
        else:
            rst = None

        body = list(body) + list(self.make_case()
                                 if case else self.make_if())
        dst_var = self.seq.dst_var
        dst_var.update(self.dst_var)
        dsts = dst_var.values()
        src_visitor = SubstSrcVisitor()

        # collect sources in destination variable definitions
        for dst in dsts:
            if isinstance(dst, (vtypes.Pointer, vtypes.Slice, vtypes.Cat)):
                raise ValueError(
                    'Partial assignment is not supported by as_module mode.')

            if isinstance(dst, vtypes._Variable):
                if dst.width is not None:
                    src_visitor.visit(dst.width)
                if dst.length is not None:
                    src_visitor.visit(dst.length)

        # collect sources in statements
        for statement in body:
            src_visitor.visit(statement)

        srcs = src_visitor.srcs.values()

        # collect sources in source variable definitions
        for src in srcs:
            if isinstance(src, vtypes._Variable):
                if src.width is not None:
                    src_visitor.visit(src.width)
                if src.length is not None:
                    src_visitor.visit(src.length)

        srcs = src_visitor.srcs.values()

        params = collections.OrderedDict()
        ports = collections.OrderedDict()
        src_rename_dict = collections.OrderedDict()

        fsm_orig_labels = [v.name for v in self.mark.values()]
        fsm_labels = collections.OrderedDict()

        # create parameter/localparam definitions
        for src in srcs:
            if isinstance(src, (vtypes.Parameter, vtypes.Localparam)):
                if src.name in fsm_orig_labels:
                    fsm_labels[src.name] = m.Localparam(src.name, src.value)
                    continue

                arg_name = src.name
                v = m.Parameter(arg_name, src.value, src.width, src.signed)
                src_rename_dict[src.name] = v
                params[arg_name] = src

        src_rename_visitor = SrcRenameVisitor(src_rename_dict)

        state_width = src_rename_visitor.visit(self.state.width)
        state_initval = src_rename_visitor.visit(self.state.initval)
        state = m.OutputReg(self.state.name, state_width,
                            initval=state_initval)
        out_state = self.m.TmpWire(state_width,
                                   prefix='_%s_out' % self.state.name)
        self.m.Always()(self.state(out_state, blk=True))
        ports[state.name] = out_state

        src_rename_dict[self.state.name] = state

        for delay, s in sorted(self.delayed_state.items(), key=lambda x: x[0]):
            s_width = src_rename_visitor.visit(s.width)
            s_initval = src_rename_visitor.visit(s.initval)
            d = m.OutputReg(s.name, s_width, initval=s_initval)
            out_d = self.m.TmpWire(s_width, prefix='_%s_out' % d.name)
            self.m.Always()(s(out_d, blk=True))
            ports[s.name] = out_d

        state_names = [self.state.name]
        state_names.extend([s.name for s in self.delayed_state.values()])

        for src in srcs:
            if isinstance(src, (vtypes.Parameter, vtypes.Localparam)):
                continue

            if src.name in state_names:
                continue

            if src.name in list(self.delayed_cond.keys()):
                rep_width = (src_rename_visitor.visit(src.width)
                             if src.width is not None else None)
                v = m.Reg(src.name, rep_width, initval=0)
                src_rename_dict[src.name] = v
                continue

            arg_name = 'i_%s' % src.name

            if src.length is not None:
                width = src.bit_length()
                length = src.length
                pack_width = vtypes.Mul(width, length)
                out_line = self.m.TmpWire(pack_width, prefix='_%s' % self.name)
                i = self.m.TmpGenvar(prefix='i')
                v = out_line[i * width:(i + 1) * width]
                g = self.m.GenerateFor(i(0), i < length, i(i + 1))
                p = g.Assign(v(src[i]))

                rep_width = (src_rename_visitor.visit(src.width)
                             if src.width is not None else None)
                rep_length = src_rename_visitor.visit(src.length)
                pack_width = (rep_length if rep_width is None else
                              vtypes.Mul(rep_length, rep_width))
                in_line = m.Input(arg_name + '_line', pack_width,
                                  signed=src.get_signed())
                in_array = m.Wire(arg_name, rep_width, rep_length,
                                  signed=src.get_signed())
                i = m.TmpGenvar(prefix='i')
                v = in_line[i * rep_width:(i + 1) * rep_width]
                g = m.GenerateFor(i(0), i < rep_length, i(i + 1))
                p = g.Assign(in_array[i](v))

                src_rename_dict[src.name] = in_array
                ports[in_line.name] = out_line

            else:
                rep_width = (src_rename_visitor.visit(src.width)
                             if src.width is not None else None)
                v = m.Input(arg_name, rep_width, signed=src.get_signed())

                src_rename_dict[src.name] = v
                ports[arg_name] = src

        for dst in dsts:
            if dst.name in list(self.delayed_cond.keys()):
                continue

            arg_name = dst.name

            rep_width = (src_rename_visitor.visit(dst.width)
                         if dst.width is not None else None)
            out = m.OutputReg(arg_name, rep_width, signed=dst.get_signed())
            out_wire = self.m.TmpWire(rep_width, signed=dst.get_signed(),
                                      prefix='_%s_%s' % (self.name, arg_name))
            self.m.Always()(dst(out_wire, blk=True))
            ports[arg_name] = out_wire

        body = [src_rename_visitor.visit(statement)
                for statement in body]

        reset = self.make_reset(reset)

        if not reset and not body:
            pass
        elif not reset or rst is None:
            m.Always(vtypes.Posedge(clk))(
                body,
            )
        else:
            m.Always(vtypes.Posedge(clk))(
                vtypes.If(rst)(
                    reset,
                )(
                    body,
                ))

        arg_params = [(name, param) for name, param in params.items()]

        arg_ports = [('CLK', self.clk)]
        if self.rst is not None:
            arg_ports.append(('RST', self.rst))

        arg_ports.extend([(name, port) for name, port in ports.items()])

        sub = Submodule(self.m, m, 'inst_' + m.name, '_%s_' % self.name,
                        arg_params=arg_params, arg_ports=arg_ports)

    # -------------------------------------------------------------------------
    def make_case(self):
        indexes = set(self.body.keys())
        indexes.update(set(self.jump.keys()))

        for index in indexes:
            self._add_mark(index)

        ret = []
        ret.extend(self.seq.make_code())
        ret.extend(self._get_delayed_substs())

        for delay, dct in sorted(self.delayed_body.items(),
                                 key=lambda x: x[0], reverse=True):
            body = tuple([self._get_delayed_when_statement(index, delay)
                          for index in sorted(dct.keys(), key=lambda x:x)])
            case = vtypes.Case(self._get_delayed_state(delay))(*body)
            ret.append(case)

        body = tuple([self._get_when_statement(index)
                      for index in sorted(indexes, key=lambda x:x)])
        case = vtypes.Case(self.state)(*body)

        if len(case.statement) > 0:
            ret.append(case)

        return ret

    def make_if(self):
        indexes = set(self.body.keys())
        indexes.update(set(self.jump.keys()))

        for index in indexes:
            self._add_mark(index)

        ret = []
        ret.extend(self.seq.make_code())
        ret.extend(self._get_delayed_substs())

        for delay, dct in sorted(self.delayed_body.items(),
                                 key=lambda x: x[0], reverse=True):
            ret.append([self._get_delayed_if_statement(index, delay)
                        for index in sorted(dct.keys(), key=lambda x:x)])

        ret.extend([self._get_if_statement(index)
                    for index in sorted(indexes, key=lambda x:x)])
        return ret

    # -------------------------------------------------------------------------
    def make_reset(self, reset):
        ret = collections.OrderedDict()

        for v in reset:
            if not isinstance(v, vtypes.Subst):
                raise TypeError(
                    'make_reset requires Subst, not %s' % str(type(v)))
            key = str(v.left)
            if key not in ret:
                ret[key] = v

        v = self.reset_visitor.visit(self.state)
        key = str(self.state)
        if v is not None and key not in ret:
            ret[key] = v

        for dst in self.delayed_state.values():
            v = self.reset_visitor.visit(dst)
            if v is None:
                continue
            key = str(v.left)
            if key not in ret:
                ret[key] = v

        for dst in self.dst_var.values():
            v = self.reset_visitor.visit(dst)
            if v is None:
                continue
            key = str(v.left)
            if key not in ret:
                ret[key] = v

        for v in self.seq.make_reset():
            if not isinstance(v, vtypes.Subst):
                raise TypeError(
                    'make_reset requires Subst, not %s' % str(type(v)))
            key = str(v.left)
            if key not in ret:
                ret[key] = v

        return list(ret.values())

    # -------------------------------------------------------------------------
    def set_index(self, index):
        return self._set_index(index)

    # -------------------------------------------------------------------------
    def _go(self, src, dst, cond=None, else_dst=None):
        self._add_jump(src, dst, cond, else_dst)
        return self

    def _add_jump(self, src, dst, cond=None, else_dst=None):
        self.jump[src].append((dst, cond, else_dst))

    # -------------------------------------------------------------------------
    def _add_statement(self, statement, index=None, keep=None, delay=None, cond=None,
                       lazy_cond=False, eager_val=False, no_delay_cond=False):

        cond = make_condition(cond)
        index = self._to_index(index) if index is not None else self.current

        if keep is not None:
            for i in range(keep):
                new_delay = i if delay is None else delay + i
                self._add_statement(statement, index=index,
                                    keep=None, delay=new_delay, cond=cond,
                                    lazy_cond=lazy_cond, eager_val=eager_val,
                                    no_delay_cond=no_delay_cond)
            return self

        if delay is not None and delay > 0:
            self._add_delayed_state(delay)

            if eager_val:
                statement = [self._add_delayed_subst(s, index, delay)
                             for s in statement]

            if not no_delay_cond:
                if cond is None:
                    cond = 1

                if not lazy_cond:
                    cond = self._add_delayed_cond(cond, index, delay)

                else:  # lazy condition
                    t = self._add_delayed_cond(1, index, delay)
                    if isinstance(cond, int) and cond == 1:
                        cond = t
                    else:
                        cond = vtypes.Ands(t, cond)

                statement = [vtypes.If(cond)(*statement)]

            self.delayed_body[delay][index].extend(statement)
            self._add_dst_var(statement)

            return self

        if cond is not None:
            statement = [vtypes.If(cond)(*statement)]
            self.last_if_statement = statement[0]

        self.body[index].extend(statement)
        self._add_dst_var(statement)

        return self

    # -------------------------------------------------------------------------
    def _add_dst_var(self, statement):
        for s in statement:
            values = self.dst_visitor.visit(s)
            for v in values:
                k = str(v)
                if k not in self.dst_var:
                    self.dst_var[k] = v

    # -------------------------------------------------------------------------
    def _add_delayed_cond(self, statement, index, delay):
        name_prefix = '_'.join(
            ['', self.name, 'cond', str(index), str(self.tmp_count)])
        self.tmp_count += 1
        prev = statement
        for i in range(delay):
            tmp_name = '_'.join([name_prefix, str(i + 1)])
            tmp = self.m.Reg(tmp_name, initval=0)
            self.delayed_cond[tmp_name] = tmp
            self._add_statement([tmp(prev)], delay=i, no_delay_cond=True)
            prev = tmp
        return prev

    # -------------------------------------------------------------------------
    def _add_delayed_subst(self, subst, index, delay):
        if not isinstance(subst, vtypes.Subst):
            return subst
        left = subst.left
        right = subst.right
        if isinstance(right, (bool, int, float, str,
                              vtypes._Constant, vtypes._ParameterVariable)):
            return subst
        width = left.bit_length()
        signed = vtypes.get_signed(left)
        prev = right

        name_prefix = ('_'.join(['', left.name, str(index), str(self.tmp_count)])
                       if isinstance(left, vtypes._Variable) else
                       '_'.join(['', self.name, 'sbst', str(index), str(self.tmp_count)]))
        self.tmp_count += 1

        for i in range(delay):
            tmp_name = '_'.join([name_prefix, str(i + 1)])
            tmp = self.m.Reg(tmp_name, width, initval=0, signed=signed)
            self._add_statement([tmp(prev)], delay=i, no_delay_cond=True)
            prev = tmp
        return left(prev)

    # -------------------------------------------------------------------------
    def _clear_next_kwargs(self):
        self.next_kwargs = {}

    def _clear_last_if_statement(self):
        self.last_if_statement = None

    def _clear_last_cond(self):
        self.last_cond = []

    def _clear_elif_cond(self):
        self.elif_cond = None

    def _make_cond(self, condlist):
        ret = None
        for cond in condlist:
            if ret is None:
                ret = cond
            else:
                ret = vtypes.Ands(ret, cond)
        return ret

    # -------------------------------------------------------------------------
    def _set_index(self, index=None):
        if index is None:
            self.state_count += 1
            return self.state_count
        self.state_count = index
        return self.state_count

    def _get_mark(self, index=None):
        if index is None:
            index = self.state_count
        if index not in self.mark:
            raise KeyError("No such index in FSM marks: %s" % index)
        return self.mark[index]

    def _set_mark(self, index=None, name=None):
        if index is None:
            index = self.state_count
        if name is None:
            name = self.name + '_' + str(index)
        self.mark[index] = self.m.Localparam(name, index)

    def _get_mark_index(self, s):
        for index, m in self.mark.items():
            if m.name == s.name:
                return index
        raise KeyError("No such mark in FSM marks: %s" % s.name)

    # -------------------------------------------------------------------------
    def _add_mark(self, index):
        index = self._to_index(index)
        if index not in self.mark:
            self._set_mark(index)
        mark = self._get_mark(index)
        return mark

    def _to_index(self, index):
        if not isinstance(index, int):
            index = self._get_mark_index(index)
        return index

    # -------------------------------------------------------------------------
    def _add_delayed_state(self, value):
        if not isinstance(value, int):
            raise TypeError("Delay amount must be int, not '%s'" %
                            str(type(value)))

        if value < 0:
            raise ValueError("Delay amount must be positive number")

        if value == 0:
            return self.state

        if value <= self.delay_amount:
            return self._get_delayed_state(value)

        for i in range(self.delay_amount + 1, value + 1):
            d = self.m.Reg(''.join(['_d', str(i), '_', self.name]), self.width,
                           initval=self._get_mark(0))
            self.delayed_state[i] = d

        self.delay_amount = value
        return d

    def _get_delayed_state(self, value):
        if value == 0:
            return self.state
        if value not in self.delayed_state:
            raise IndexError('No such index %d in delayed state' % value)
        return self.delayed_state[value]

    def _get_delayed_substs(self):
        ret = []
        prev = self.state
        for d in range(1, self.delay_amount + 1):
            ret.append(vtypes.Subst(self.delayed_state[d], prev))
            prev = self.delayed_state[d]
        return ret

    def _init_delayed_state(self):
        ret = []
        for d in range(1, self.delay_amount + 1):
            ret.append(vtypes.Subst(self.delayed_state[d], self._get_mark(0)))
        return ret

    def _to_state_assign(self, dst, cond=None, else_dst=None):
        dst_mark = self._get_mark(dst)
        value = self.state(dst_mark)
        if cond is not None:
            value = vtypes.If(cond)(value)
            if else_dst is not None:
                else_dst_mark = self._get_mark(else_dst)
                value = value.Else(self.state(else_dst_mark))
        return value

    # -------------------------------------------------------------------------
    def _cond_case(self, index):
        if index not in self.mark:
            self._set_mark(index)
        return self._get_mark(index)

    def _cond_if(self, index):
        if index not in self.mark:
            self._set_mark(index)
        return (self.state == self._get_mark(index))

    def _delayed_cond_if(self, index, delay):
        if index not in self.mark:
            self._set_mark(index)
        if delay > 0 and delay not in self.delayed_state:
            self._add_delayed_state(delay)
        return (self._get_delayed_state(delay) == self._get_mark(index))

    def _get_when_statement(self, index):
        body = []
        body.extend(self.body[index])
        for dst, cond, else_dst in self.jump[index]:
            self._add_mark(dst)
            if else_dst is not None:
                self._add_mark(else_dst)
            body.append(self._to_state_assign(dst, cond, else_dst))
        return vtypes.When(self._cond_case(index))(*body)

    def _get_delayed_when_statement(self, index, delay):
        return vtypes.When(self._cond_case(index))(*self.delayed_body[delay][index])

    def _get_if_statement(self, index):
        body = []
        body.extend(self.body[index])
        for dst, cond, else_dst in self.jump[index]:
            self._add_mark(dst)
            if else_dst is not None:
                self._add_mark(else_dst)
            body.append(self._to_state_assign(dst, cond, else_dst))
        return vtypes.If(self._cond_if(index))(*body)

    def _get_delayed_if_statement(self, index, delay):
        return vtypes.If(self._delayed_cond_if(index, delay))(*self.delayed_body[delay][index])

    # -------------------------------------------------------------------------
    def __call__(self, *statement, **kwargs):
        return self.add(*statement, **kwargs)

    def __getitem__(self, index):
        return self.body[index]

    def __len__(self):
        return self.state_count + 1



from __future__ import absolute_import
from __future__ import print_function
import sys
import os

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.abspath(__file__)))))

from veriloggen import *
import veriloggen.thread as vthread
import veriloggen.types.axi as axi
import veriloggen.types.ipcore as ipcore


def mkLed():
    m = Module('add')
    clk = m.Input('CLK')
    rst = m.Input('RST')

    saxi = vthread.AXISLiteRegister(m, 'saxi', clk, rst,
                                    datawidth=32, length=8)

    def add():
        while True:
            saxi.wait_flag(0, value=1, resetvalue=0)

            a = saxi.read(2)
            b = saxi.read(3)
            c = a + b

            saxi.write(4, c)
            saxi.write_flag(1, 1, resetvalue=0)

    th = vthread.Thread(m, 'th_add', clk, rst, add)
    fsm = th.start()

    return m


def mkTest():
    m = Module('test')

    # target instance
    led = mkLed()

    # copy paras and ports
    params = m.copy_params(led)
    ports = m.copy_sim_ports(led)

    clk = ports['CLK']
    rst = ports['RST']

    # AXI-Slave controller
    _saxi = vthread.AXIMLite(m, '_saxi', clk, rst, noio=True)
    _saxi.connect(ports, 'saxi')

    # Timer
    counter = m.Reg('counter', 32, initval=0)
    seq = Seq(m, 'seq', clk, rst)
    seq(
        counter.inc()
    )

    def ctrl():
        for i in range(100):
            pass

        awaddr = 2 * 4
        a = 10
        print('# a = %d' % a)
        _saxi.write(awaddr, a)

        awaddr = 3 * 4
        b = 20
        print('# b = %d' % b)
        _saxi.write(awaddr, b)

        awaddr = 0 * 4
        start_time = counter
        print('# start time = %d' % start_time)
        _saxi.write(awaddr, 1)

        araddr = 1 * 4
        v = _saxi.read(araddr)
        while v == 0:
            v = _saxi.read(araddr)

        araddr = 4 * 4
        c = _saxi.read(araddr)
        print('# c = %d' % c)

        end_time = counter
        print('# end time = %d' % end_time)
        time = end_time - start_time
        print('# exec time = %d' % time)

    th = vthread.Thread(m, 'th_ctrl', clk, rst, ctrl)
    fsm = th.start()

    uut = m.Instance(led, 'uut',
                     params=m.connect_params(led),
                     ports=m.connect_ports(led))

    simulation.setup_waveform(m, uut)
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(100000),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    m = mkLed()
    ipcore.to_ipcore(m, iftype='axi')



from __future__ import absolute_import
from __future__ import print_function

import veriloggen.core.vtypes as vtypes
import veriloggen.core.module as module


def mkMaddCore(index, awidth=32, bwidth=32, cwidth=32,
               asigned=True, bsigned=True, csigned=True, depth=6):

    retwidth = max(awidth + bwidth, cwidth)

    m = module.Module('madd_core_%d' % index)

    clk = m.Input('CLK')
    update = m.Input('update')

    a = m.Input('a', awidth)
    b = m.Input('b', bwidth)
    c = m.Input('c', cwidth)
    d = m.Output('d', retwidth)

    _a = m.Reg('_a', awidth, signed=asigned)
    _b = m.Reg('_b', bwidth, signed=bsigned)
    _c = m.Reg('_c', cwidth, signed=csigned)
    _mul = m.Wire('_mul', retwidth, signed=True)
    _madd = m.Wire('_madd', retwidth, signed=True)
    _pipe_madd = [m.Reg('_pipe_madd%d' % i, retwidth, signed=True)
                  for i in range(depth - 1)]

    __a = _a
    __b = _b
    __c = _c
    if not asigned:
        __a = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _a))
    if not bsigned:
        __b = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _b))
    if not csigned:
        __c = vtypes.SystemTask(
            'signed', vtypes.Cat(vtypes.Int(0, width=1), _c))

    m.Assign(_mul(__a * __b))
    m.Assign(_madd(_mul + __c))
    m.Assign(d(_pipe_madd[depth - 2]))

    m.Always(vtypes.Posedge(clk))(
        vtypes.If(update)(
            _a(a),
            _b(b),
            _c(c),
            _pipe_madd[0](_madd),
            [_pipe_madd[i](_pipe_madd[i - 1]) for i in range(1, depth - 1)]
        ))

    return m


def mkMadd(index, awidth=32, bwidth=32, cwidth=32,
           asigned=True, bsigned=True, csigned=True, depth=6):

    if awidth < 0:
        raise ValueError("data width must be greater than 0.")
    if bwidth < 0:
        raise ValueError("data width must be greater than 0.")
    if cwidth < 0:
        raise ValueError("data width must be greater than 0.")
    if depth < 2:
        raise ValueError("depth must be greater than 2.")

    retwidth = max(awidth + bwidth, cwidth)

    madd = mkMaddCore(index, awidth, bwidth, cwidth,
                      asigned, bsigned, csigned, depth)

    m = module.Module('madd_%d' % index)

    clk = m.Input('CLK')
    update = m.Input('update')
    a = m.Input('a', awidth)
    b = m.Input('b', bwidth)
    c = m.Input('c', cwidth)
    d = m.Output('d', retwidth)

    ports = [('CLK', clk), ('update', update),
             ('a', a), ('b', b), ('c', c), ('d', d)]
    m.Instance(madd, 'madd', ports=ports)

    return m


# global madd count
index_count = 0


def get_madd(awidth=32, bwidth=32, cwidth=32,
             asigned=True, bsigned=True, csigned=True, depth=6):

    global index_count
    madd = mkMadd(index_count, awidth, bwidth, cwidth,
                  asigned, bsigned, csigned, depth)
    index_count += 1
    return madd


def reset():
    global index_count
    index_count = 0



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import thread_fifo_multithread

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  wire [8-1:0] LED;

  blinkled
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .LED(LED)
  );


  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #10000;
    $finish;
  end


endmodule



module blinkled
(
  input CLK,
  input RST,
  output reg [8-1:0] LED
);

  reg myfifo_enq;
  reg [32-1:0] myfifo_wdata;
  wire myfifo_full;
  wire myfifo_almost_full;
  reg myfifo_deq;
  wire [32-1:0] myfifo_rdata;
  wire myfifo_empty;
  wire myfifo_almost_empty;

  myfifo
  inst_myfifo
  (
    .CLK(CLK),
    .RST(RST),
    .myfifo_enq(myfifo_enq),
    .myfifo_wdata(myfifo_wdata),
    .myfifo_full(myfifo_full),
    .myfifo_almost_full(myfifo_almost_full),
    .myfifo_deq(myfifo_deq),
    .myfifo_rdata(myfifo_rdata),
    .myfifo_empty(myfifo_empty),
    .myfifo_almost_empty(myfifo_almost_empty)
  );

  reg [5-1:0] count_myfifo;
  reg [8-1:0] _th_myfunc_start;
  reg [32-1:0] th_blink;
  localparam th_blink_init = 0;
  reg _myfifo_cond_0_1;
  reg _myfifo_cond_1_1;
  reg signed [32-1:0] _th_blink_tid_0;
  reg [32-1:0] th_myfunc_0;
  localparam th_myfunc_0_init = 0;
  reg [32-1:0] th_myfunc_1;
  localparam th_myfunc_1_init = 0;
  reg [32-1:0] th_myfunc_2;
  localparam th_myfunc_2_init = 0;
  reg [32-1:0] th_myfunc_3;
  localparam th_myfunc_3_init = 0;
  reg [32-1:0] th_myfunc_4;
  localparam th_myfunc_4_init = 0;
  reg [32-1:0] th_myfunc_5;
  localparam th_myfunc_5_init = 0;
  reg [32-1:0] th_myfunc_6;
  localparam th_myfunc_6_init = 0;
  reg [32-1:0] th_myfunc_7;
  localparam th_myfunc_7_init = 0;
  reg _th_myfunc_0_called;
  reg signed [32-1:0] _th_myfunc_0_tid_1;
  reg signed [32-1:0] _th_myfunc_0_tid_2;
  reg __myfifo_mutex_lock_reg;
  reg [32-1:0] __myfifo_mutex_lock_id;
  reg _tmp_0;
  reg _myfifo_cond_2_1;
  reg _myfifo_cond_3_1;
  reg _myfifo_cond_4_1;
  reg _myfifo_cond_4_2;
  reg signed [32-1:0] _tmp_1;
  reg signed [32-1:0] _th_myfunc_0_read_data_3;
  reg signed [32-1:0] _th_myfunc_0_write_data_4;
  reg _myfifo_cond_5_1;
  reg _tmp_2;
  reg _myfifo_cond_6_1;
  reg _myfifo_cond_7_1;
  reg _myfifo_cond_8_1;
  reg _myfifo_cond_8_2;
  reg signed [32-1:0] _tmp_3;
  reg _myfifo_cond_9_1;
  reg _th_myfunc_1_called;
  reg signed [32-1:0] _th_myfunc_1_tid_5;
  reg signed [32-1:0] _th_myfunc_1_tid_6;
  reg _tmp_4;
  reg _myfifo_cond_10_1;
  reg _myfifo_cond_11_1;
  reg _myfifo_cond_12_1;
  reg _myfifo_cond_12_2;
  reg signed [32-1:0] _tmp_5;
  reg signed [32-1:0] _th_myfunc_1_read_data_7;
  reg signed [32-1:0] _th_myfunc_1_write_data_8;
  reg _myfifo_cond_13_1;
  reg _tmp_6;
  reg _myfifo_cond_14_1;
  reg _myfifo_cond_15_1;
  reg _myfifo_cond_16_1;
  reg _myfifo_cond_16_2;
  reg signed [32-1:0] _tmp_7;
  reg _myfifo_cond_17_1;
  reg _th_myfunc_2_called;
  reg signed [32-1:0] _th_myfunc_2_tid_9;
  reg signed [32-1:0] _th_myfunc_2_tid_10;
  reg _tmp_8;
  reg _myfifo_cond_18_1;
  reg _myfifo_cond_19_1;
  reg _myfifo_cond_20_1;
  reg _myfifo_cond_20_2;
  reg signed [32-1:0] _tmp_9;
  reg signed [32-1:0] _th_myfunc_2_read_data_11;
  reg signed [32-1:0] _th_myfunc_2_write_data_12;
  reg _myfifo_cond_21_1;
  reg _tmp_10;
  reg _myfifo_cond_22_1;
  reg _myfifo_cond_23_1;
  reg _myfifo_cond_24_1;
  reg _myfifo_cond_24_2;
  reg signed [32-1:0] _tmp_11;
  reg _myfifo_cond_25_1;
  reg _th_myfunc_3_called;
  reg signed [32-1:0] _th_myfunc_3_tid_13;
  reg signed [32-1:0] _th_myfunc_3_tid_14;
  reg _tmp_12;
  reg _myfifo_cond_26_1;
  reg _myfifo_cond_27_1;
  reg _myfifo_cond_28_1;
  reg _myfifo_cond_28_2;
  reg signed [32-1:0] _tmp_13;
  reg signed [32-1:0] _th_myfunc_3_read_data_15;
  reg signed [32-1:0] _th_myfunc_3_write_data_16;
  reg _myfifo_cond_29_1;
  reg _tmp_14;
  reg _myfifo_cond_30_1;
  reg _myfifo_cond_31_1;
  reg _myfifo_cond_32_1;
  reg _myfifo_cond_32_2;
  reg signed [32-1:0] _tmp_15;
  reg _myfifo_cond_33_1;
  reg _th_myfunc_4_called;
  reg signed [32-1:0] _th_myfunc_4_tid_17;
  reg signed [32-1:0] _th_myfunc_4_tid_18;
  reg _tmp_16;
  reg _myfifo_cond_34_1;
  reg _myfifo_cond_35_1;
  reg _myfifo_cond_36_1;
  reg _myfifo_cond_36_2;
  reg signed [32-1:0] _tmp_17;
  reg signed [32-1:0] _th_myfunc_4_read_data_19;
  reg signed [32-1:0] _th_myfunc_4_write_data_20;
  reg _myfifo_cond_37_1;
  reg _tmp_18;
  reg _myfifo_cond_38_1;
  reg _myfifo_cond_39_1;
  reg _myfifo_cond_40_1;
  reg _myfifo_cond_40_2;
  reg signed [32-1:0] _tmp_19;
  reg _myfifo_cond_41_1;
  reg _th_myfunc_5_called;
  reg signed [32-1:0] _th_myfunc_5_tid_21;
  reg signed [32-1:0] _th_myfunc_5_tid_22;
  reg _tmp_20;
  reg _myfifo_cond_42_1;
  reg _myfifo_cond_43_1;
  reg _myfifo_cond_44_1;
  reg _myfifo_cond_44_2;
  reg signed [32-1:0] _tmp_21;
  reg signed [32-1:0] _th_myfunc_5_read_data_23;
  reg signed [32-1:0] _th_myfunc_5_write_data_24;
  reg _myfifo_cond_45_1;
  reg _tmp_22;
  reg _myfifo_cond_46_1;
  reg _myfifo_cond_47_1;
  reg _myfifo_cond_48_1;
  reg _myfifo_cond_48_2;
  reg signed [32-1:0] _tmp_23;
  reg _myfifo_cond_49_1;
  reg _th_myfunc_6_called;
  reg signed [32-1:0] _th_myfunc_6_tid_25;
  reg signed [32-1:0] _th_myfunc_6_tid_26;
  reg _tmp_24;
  reg _myfifo_cond_50_1;
  reg _myfifo_cond_51_1;
  reg _myfifo_cond_52_1;
  reg _myfifo_cond_52_2;
  reg signed [32-1:0] _tmp_25;
  reg signed [32-1:0] _th_myfunc_6_read_data_27;
  reg signed [32-1:0] _th_myfunc_6_write_data_28;
  reg _myfifo_cond_53_1;
  reg _tmp_26;
  reg _myfifo_cond_54_1;
  reg _myfifo_cond_55_1;
  reg _myfifo_cond_56_1;
  reg _myfifo_cond_56_2;
  reg signed [32-1:0] _tmp_27;
  reg _myfifo_cond_57_1;
  reg _th_myfunc_7_called;
  reg signed [32-1:0] _th_myfunc_7_tid_29;
  reg signed [32-1:0] _th_myfunc_7_tid_30;
  reg _tmp_28;
  reg _myfifo_cond_58_1;
  reg _myfifo_cond_59_1;
  reg _myfifo_cond_60_1;
  reg _myfifo_cond_60_2;
  reg signed [32-1:0] _tmp_29;
  reg signed [32-1:0] _th_myfunc_7_read_data_31;
  reg signed [32-1:0] _th_myfunc_7_write_data_32;
  reg _myfifo_cond_61_1;
  reg _tmp_30;
  reg _myfifo_cond_62_1;
  reg _myfifo_cond_63_1;
  reg _myfifo_cond_64_1;
  reg _myfifo_cond_64_2;
  reg signed [32-1:0] _tmp_31;
  reg _myfifo_cond_65_1;
  reg _tmp_32;
  reg _myfifo_cond_66_1;
  reg _myfifo_cond_67_1;
  reg _myfifo_cond_68_1;
  reg _myfifo_cond_68_2;
  reg signed [32-1:0] _tmp_33;
  reg signed [32-1:0] _th_blink_read_data_33;
  reg _tmp_34;
  reg _myfifo_cond_69_1;
  reg _myfifo_cond_70_1;
  reg _myfifo_cond_71_1;
  reg _myfifo_cond_71_2;
  reg signed [32-1:0] _tmp_35;

  always @(posedge CLK) begin
    if(RST) begin
      count_myfifo <= 0;
      myfifo_wdata <= 0;
      myfifo_enq <= 0;
      _myfifo_cond_0_1 <= 0;
      _myfifo_cond_1_1 <= 0;
      myfifo_deq <= 0;
      _myfifo_cond_2_1 <= 0;
      _tmp_0 <= 0;
      _myfifo_cond_3_1 <= 0;
      _myfifo_cond_4_1 <= 0;
      _myfifo_cond_4_2 <= 0;
      _myfifo_cond_5_1 <= 0;
      _myfifo_cond_6_1 <= 0;
      _tmp_2 <= 0;
      _myfifo_cond_7_1 <= 0;
      _myfifo_cond_8_1 <= 0;
      _myfifo_cond_8_2 <= 0;
      _myfifo_cond_9_1 <= 0;
      _myfifo_cond_10_1 <= 0;
      _tmp_4 <= 0;
      _myfifo_cond_11_1 <= 0;
      _myfifo_cond_12_1 <= 0;
      _myfifo_cond_12_2 <= 0;
      _myfifo_cond_13_1 <= 0;
      _myfifo_cond_14_1 <= 0;
      _tmp_6 <= 0;
      _myfifo_cond_15_1 <= 0;
      _myfifo_cond_16_1 <= 0;
      _myfifo_cond_16_2 <= 0;
      _myfifo_cond_17_1 <= 0;
      _myfifo_cond_18_1 <= 0;
      _tmp_8 <= 0;
      _myfifo_cond_19_1 <= 0;
      _myfifo_cond_20_1 <= 0;
      _myfifo_cond_20_2 <= 0;
      _myfifo_cond_21_1 <= 0;
      _myfifo_cond_22_1 <= 0;
      _tmp_10 <= 0;
      _myfifo_cond_23_1 <= 0;
      _myfifo_cond_24_1 <= 0;
      _myfifo_cond_24_2 <= 0;
      _myfifo_cond_25_1 <= 0;
      _myfifo_cond_26_1 <= 0;
      _tmp_12 <= 0;
      _myfifo_cond_27_1 <= 0;
      _myfifo_cond_28_1 <= 0;
      _myfifo_cond_28_2 <= 0;
      _myfifo_cond_29_1 <= 0;
      _myfifo_cond_30_1 <= 0;
      _tmp_14 <= 0;
      _myfifo_cond_31_1 <= 0;
      _myfifo_cond_32_1 <= 0;
      _myfifo_cond_32_2 <= 0;
      _myfifo_cond_33_1 <= 0;
      _myfifo_cond_34_1 <= 0;
      _tmp_16 <= 0;
      _myfifo_cond_35_1 <= 0;
      _myfifo_cond_36_1 <= 0;
      _myfifo_cond_36_2 <= 0;
      _myfifo_cond_37_1 <= 0;
      _myfifo_cond_38_1 <= 0;
      _tmp_18 <= 0;
      _myfifo_cond_39_1 <= 0;
      _myfifo_cond_40_1 <= 0;
      _myfifo_cond_40_2 <= 0;
      _myfifo_cond_41_1 <= 0;
      _myfifo_cond_42_1 <= 0;
      _tmp_20 <= 0;
      _myfifo_cond_43_1 <= 0;
      _myfifo_cond_44_1 <= 0;
      _myfifo_cond_44_2 <= 0;
      _myfifo_cond_45_1 <= 0;
      _myfifo_cond_46_1 <= 0;
      _tmp_22 <= 0;
      _myfifo_cond_47_1 <= 0;
      _myfifo_cond_48_1 <= 0;
      _myfifo_cond_48_2 <= 0;
      _myfifo_cond_49_1 <= 0;
      _myfifo_cond_50_1 <= 0;
      _tmp_24 <= 0;
      _myfifo_cond_51_1 <= 0;
      _myfifo_cond_52_1 <= 0;
      _myfifo_cond_52_2 <= 0;
      _myfifo_cond_53_1 <= 0;
      _myfifo_cond_54_1 <= 0;
      _tmp_26 <= 0;
      _myfifo_cond_55_1 <= 0;
      _myfifo_cond_56_1 <= 0;
      _myfifo_cond_56_2 <= 0;
      _myfifo_cond_57_1 <= 0;
      _myfifo_cond_58_1 <= 0;
      _tmp_28 <= 0;
      _myfifo_cond_59_1 <= 0;
      _myfifo_cond_60_1 <= 0;
      _myfifo_cond_60_2 <= 0;
      _myfifo_cond_61_1 <= 0;
      _myfifo_cond_62_1 <= 0;
      _tmp_30 <= 0;
      _myfifo_cond_63_1 <= 0;
      _myfifo_cond_64_1 <= 0;
      _myfifo_cond_64_2 <= 0;
      _myfifo_cond_65_1 <= 0;
      _myfifo_cond_66_1 <= 0;
      _tmp_32 <= 0;
      _myfifo_cond_67_1 <= 0;
      _myfifo_cond_68_1 <= 0;
      _myfifo_cond_68_2 <= 0;
      _myfifo_cond_69_1 <= 0;
      _tmp_34 <= 0;
      _myfifo_cond_70_1 <= 0;
      _myfifo_cond_71_1 <= 0;
      _myfifo_cond_71_2 <= 0;
    end else begin
      if(_myfifo_cond_4_2) begin
        _tmp_0 <= 0;
      end 
      if(_myfifo_cond_8_2) begin
        _tmp_2 <= 0;
      end 
      if(_myfifo_cond_12_2) begin
        _tmp_4 <= 0;
      end 
      if(_myfifo_cond_16_2) begin
        _tmp_6 <= 0;
      end 
      if(_myfifo_cond_20_2) begin
        _tmp_8 <= 0;
      end 
      if(_myfifo_cond_24_2) begin
        _tmp_10 <= 0;
      end 
      if(_myfifo_cond_28_2) begin
        _tmp_12 <= 0;
      end 
      if(_myfifo_cond_32_2) begin
        _tmp_14 <= 0;
      end 
      if(_myfifo_cond_36_2) begin
        _tmp_16 <= 0;
      end 
      if(_myfifo_cond_40_2) begin
        _tmp_18 <= 0;
      end 
      if(_myfifo_cond_44_2) begin
        _tmp_20 <= 0;
      end 
      if(_myfifo_cond_48_2) begin
        _tmp_22 <= 0;
      end 
      if(_myfifo_cond_52_2) begin
        _tmp_24 <= 0;
      end 
      if(_myfifo_cond_56_2) begin
        _tmp_26 <= 0;
      end 
      if(_myfifo_cond_60_2) begin
        _tmp_28 <= 0;
      end 
      if(_myfifo_cond_64_2) begin
        _tmp_30 <= 0;
      end 
      if(_myfifo_cond_68_2) begin
        _tmp_32 <= 0;
      end 
      if(_myfifo_cond_71_2) begin
        _tmp_34 <= 0;
      end 
      if(_myfifo_cond_0_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_1_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_2_1) begin
        _tmp_0 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_3_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_4_2 <= _myfifo_cond_4_1;
      if(_myfifo_cond_5_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_6_1) begin
        _tmp_2 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_7_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_8_2 <= _myfifo_cond_8_1;
      if(_myfifo_cond_9_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_10_1) begin
        _tmp_4 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_11_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_12_2 <= _myfifo_cond_12_1;
      if(_myfifo_cond_13_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_14_1) begin
        _tmp_6 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_15_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_16_2 <= _myfifo_cond_16_1;
      if(_myfifo_cond_17_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_18_1) begin
        _tmp_8 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_19_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_20_2 <= _myfifo_cond_20_1;
      if(_myfifo_cond_21_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_22_1) begin
        _tmp_10 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_23_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_24_2 <= _myfifo_cond_24_1;
      if(_myfifo_cond_25_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_26_1) begin
        _tmp_12 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_27_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_28_2 <= _myfifo_cond_28_1;
      if(_myfifo_cond_29_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_30_1) begin
        _tmp_14 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_31_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_32_2 <= _myfifo_cond_32_1;
      if(_myfifo_cond_33_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_34_1) begin
        _tmp_16 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_35_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_36_2 <= _myfifo_cond_36_1;
      if(_myfifo_cond_37_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_38_1) begin
        _tmp_18 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_39_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_40_2 <= _myfifo_cond_40_1;
      if(_myfifo_cond_41_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_42_1) begin
        _tmp_20 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_43_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_44_2 <= _myfifo_cond_44_1;
      if(_myfifo_cond_45_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_46_1) begin
        _tmp_22 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_47_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_48_2 <= _myfifo_cond_48_1;
      if(_myfifo_cond_49_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_50_1) begin
        _tmp_24 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_51_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_52_2 <= _myfifo_cond_52_1;
      if(_myfifo_cond_53_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_54_1) begin
        _tmp_26 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_55_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_56_2 <= _myfifo_cond_56_1;
      if(_myfifo_cond_57_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_58_1) begin
        _tmp_28 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_59_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_60_2 <= _myfifo_cond_60_1;
      if(_myfifo_cond_61_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_62_1) begin
        _tmp_30 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_63_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_64_2 <= _myfifo_cond_64_1;
      if(_myfifo_cond_65_1) begin
        myfifo_enq <= 0;
      end 
      if(_myfifo_cond_66_1) begin
        _tmp_32 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_67_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_68_2 <= _myfifo_cond_68_1;
      if(_myfifo_cond_69_1) begin
        _tmp_34 <= !myfifo_empty && myfifo_deq;
      end 
      if(_myfifo_cond_70_1) begin
        myfifo_deq <= 0;
      end 
      _myfifo_cond_71_2 <= _myfifo_cond_71_1;
      if(myfifo_enq && !myfifo_full && (myfifo_deq && !myfifo_empty)) begin
        count_myfifo <= count_myfifo;
      end else if(myfifo_enq && !myfifo_full) begin
        count_myfifo <= count_myfifo + 1;
      end else if(myfifo_deq && !myfifo_empty) begin
        count_myfifo <= count_myfifo - 1;
      end 
      if((th_blink == 1) && !myfifo_full) begin
        myfifo_wdata <= 100;
      end 
      if((th_blink == 1) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_0_1 <= 1;
      if((th_blink == 2) && !myfifo_full) begin
        myfifo_wdata <= 200;
      end 
      if((th_blink == 2) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_1_1 <= 1;
      if(th_myfunc_0 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_2_1 <= th_myfunc_0 == 5;
      _myfifo_cond_3_1 <= 1;
      _myfifo_cond_4_1 <= 1;
      if((th_myfunc_0 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_0_write_data_4;
      end 
      if((th_myfunc_0 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_5_1 <= 1;
      if(th_myfunc_0 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_6_1 <= th_myfunc_0 == 13;
      _myfifo_cond_7_1 <= 1;
      _myfifo_cond_8_1 <= 1;
      if((th_myfunc_0 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_0_write_data_4;
      end 
      if((th_myfunc_0 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_9_1 <= 1;
      if(th_myfunc_1 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_10_1 <= th_myfunc_1 == 5;
      _myfifo_cond_11_1 <= 1;
      _myfifo_cond_12_1 <= 1;
      if((th_myfunc_1 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_1_write_data_8;
      end 
      if((th_myfunc_1 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_13_1 <= 1;
      if(th_myfunc_1 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_14_1 <= th_myfunc_1 == 13;
      _myfifo_cond_15_1 <= 1;
      _myfifo_cond_16_1 <= 1;
      if((th_myfunc_1 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_1_write_data_8;
      end 
      if((th_myfunc_1 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_17_1 <= 1;
      if(th_myfunc_2 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_18_1 <= th_myfunc_2 == 5;
      _myfifo_cond_19_1 <= 1;
      _myfifo_cond_20_1 <= 1;
      if((th_myfunc_2 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_2_write_data_12;
      end 
      if((th_myfunc_2 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_21_1 <= 1;
      if(th_myfunc_2 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_22_1 <= th_myfunc_2 == 13;
      _myfifo_cond_23_1 <= 1;
      _myfifo_cond_24_1 <= 1;
      if((th_myfunc_2 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_2_write_data_12;
      end 
      if((th_myfunc_2 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_25_1 <= 1;
      if(th_myfunc_3 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_26_1 <= th_myfunc_3 == 5;
      _myfifo_cond_27_1 <= 1;
      _myfifo_cond_28_1 <= 1;
      if((th_myfunc_3 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_3_write_data_16;
      end 
      if((th_myfunc_3 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_29_1 <= 1;
      if(th_myfunc_3 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_30_1 <= th_myfunc_3 == 13;
      _myfifo_cond_31_1 <= 1;
      _myfifo_cond_32_1 <= 1;
      if((th_myfunc_3 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_3_write_data_16;
      end 
      if((th_myfunc_3 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_33_1 <= 1;
      if(th_myfunc_4 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_34_1 <= th_myfunc_4 == 5;
      _myfifo_cond_35_1 <= 1;
      _myfifo_cond_36_1 <= 1;
      if((th_myfunc_4 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_4_write_data_20;
      end 
      if((th_myfunc_4 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_37_1 <= 1;
      if(th_myfunc_4 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_38_1 <= th_myfunc_4 == 13;
      _myfifo_cond_39_1 <= 1;
      _myfifo_cond_40_1 <= 1;
      if((th_myfunc_4 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_4_write_data_20;
      end 
      if((th_myfunc_4 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_41_1 <= 1;
      if(th_myfunc_5 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_42_1 <= th_myfunc_5 == 5;
      _myfifo_cond_43_1 <= 1;
      _myfifo_cond_44_1 <= 1;
      if((th_myfunc_5 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_5_write_data_24;
      end 
      if((th_myfunc_5 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_45_1 <= 1;
      if(th_myfunc_5 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_46_1 <= th_myfunc_5 == 13;
      _myfifo_cond_47_1 <= 1;
      _myfifo_cond_48_1 <= 1;
      if((th_myfunc_5 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_5_write_data_24;
      end 
      if((th_myfunc_5 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_49_1 <= 1;
      if(th_myfunc_6 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_50_1 <= th_myfunc_6 == 5;
      _myfifo_cond_51_1 <= 1;
      _myfifo_cond_52_1 <= 1;
      if((th_myfunc_6 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_6_write_data_28;
      end 
      if((th_myfunc_6 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_53_1 <= 1;
      if(th_myfunc_6 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_54_1 <= th_myfunc_6 == 13;
      _myfifo_cond_55_1 <= 1;
      _myfifo_cond_56_1 <= 1;
      if((th_myfunc_6 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_6_write_data_28;
      end 
      if((th_myfunc_6 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_57_1 <= 1;
      if(th_myfunc_7 == 5) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_58_1 <= th_myfunc_7 == 5;
      _myfifo_cond_59_1 <= 1;
      _myfifo_cond_60_1 <= 1;
      if((th_myfunc_7 == 11) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_7_write_data_32;
      end 
      if((th_myfunc_7 == 11) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_61_1 <= 1;
      if(th_myfunc_7 == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_62_1 <= th_myfunc_7 == 13;
      _myfifo_cond_63_1 <= 1;
      _myfifo_cond_64_1 <= 1;
      if((th_myfunc_7 == 19) && !myfifo_full) begin
        myfifo_wdata <= _th_myfunc_7_write_data_32;
      end 
      if((th_myfunc_7 == 19) && !myfifo_full) begin
        myfifo_enq <= 1;
      end 
      _myfifo_cond_65_1 <= 1;
      if(th_blink == 13) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_66_1 <= th_blink == 13;
      _myfifo_cond_67_1 <= 1;
      _myfifo_cond_68_1 <= 1;
      if(th_blink == 19) begin
        myfifo_deq <= 1;
      end 
      _myfifo_cond_69_1 <= th_blink == 19;
      _myfifo_cond_70_1 <= 1;
      _myfifo_cond_71_1 <= 1;
    end
  end

  localparam th_blink_1 = 1;
  localparam th_blink_2 = 2;
  localparam th_blink_3 = 3;
  localparam th_blink_4 = 4;
  localparam th_blink_5 = 5;
  localparam th_blink_6 = 6;
  localparam th_blink_7 = 7;
  localparam th_blink_8 = 8;
  localparam th_blink_9 = 9;
  localparam th_blink_10 = 10;
  localparam th_blink_11 = 11;
  localparam th_blink_12 = 12;
  localparam th_blink_13 = 13;
  localparam th_blink_14 = 14;
  localparam th_blink_15 = 15;
  localparam th_blink_16 = 16;
  localparam th_blink_17 = 17;
  localparam th_blink_18 = 18;
  localparam th_blink_19 = 19;
  localparam th_blink_20 = 20;
  localparam th_blink_21 = 21;
  localparam th_blink_22 = 22;
  localparam th_blink_23 = 23;
  localparam th_blink_24 = 24;
  localparam th_blink_25 = 25;

  always @(posedge CLK) begin
    if(RST) begin
      th_blink <= th_blink_init;
      _th_blink_tid_0 <= 0;
      _th_myfunc_start[_th_blink_tid_0] <= (0 >> _th_blink_tid_0) & 1'd1;
      _tmp_33 <= 0;
      _th_blink_read_data_33 <= 0;
      _tmp_35 <= 0;
    end else begin
      case(th_blink)
        th_blink_init: begin
          th_blink <= th_blink_1;
        end
        th_blink_1: begin
          if(!myfifo_almost_full) begin
            th_blink <= th_blink_2;
          end 
        end
        th_blink_2: begin
          if(!myfifo_almost_full) begin
            th_blink <= th_blink_3;
          end 
        end
        th_blink_3: begin
          _th_blink_tid_0 <= 0;
          th_blink <= th_blink_4;
        end
        th_blink_4: begin
          if(_th_blink_tid_0 < 8) begin
            th_blink <= th_blink_5;
          end else begin
            th_blink <= th_blink_9;
          end
        end
        th_blink_5: begin
          _th_myfunc_start[_th_blink_tid_0] <= 1;
          th_blink <= th_blink_6;
        end
        th_blink_6: begin
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
          th_blink <= th_blink_7;
        end
        th_blink_7: begin
          _th_myfunc_start[_th_blink_tid_0] <= 0;
          th_blink <= th_blink_8;
        end
        th_blink_8: begin
          _th_blink_tid_0 <= _th_blink_tid_0 + 1;
          th_blink <= th_blink_4;
        end
        th_blink_9: begin
          _th_blink_tid_0 <= 0;
          th_blink <= th_blink_10;
        end
        th_blink_10: begin
          if(_th_blink_tid_0 < 8) begin
            th_blink <= th_blink_11;
          end else begin
            th_blink <= th_blink_13;
          end
        end
        th_blink_11: begin
          if((_th_blink_tid_0 == 0)? th_myfunc_0 == 23 : 
          (_th_blink_tid_0 == 1)? th_myfunc_1 == 23 : 
          (_th_blink_tid_0 == 2)? th_myfunc_2 == 23 : 
          (_th_blink_tid_0 == 3)? th_myfunc_3 == 23 : 
          (_th_blink_tid_0 == 4)? th_myfunc_4 == 23 : 
          (_th_blink_tid_0 == 5)? th_myfunc_5 == 23 : 
          (_th_blink_tid_0 == 6)? th_myfunc_6 == 23 : 
          (_th_blink_tid_0 == 7)? th_myfunc_7 == 23 : 0) begin
            th_blink <= th_blink_12;
          end 
        end
        th_blink_12: begin
          _th_blink_tid_0 <= _th_blink_tid_0 + 1;
          th_blink <= th_blink_10;
        end
        th_blink_13: begin
          if(!myfifo_empty) begin
            th_blink <= th_blink_14;
          end 
        end
        th_blink_14: begin
          th_blink <= th_blink_15;
        end
        th_blink_15: begin
          if(_tmp_32) begin
            _tmp_33 <= myfifo_rdata;
          end 
          if(_tmp_32) begin
            th_blink <= th_blink_16;
          end 
        end
        th_blink_16: begin
          _th_blink_read_data_33 <= _tmp_33;
          th_blink <= th_blink_17;
        end
        th_blink_17: begin
          LED <= _th_blink_read_data_33;
          th_blink <= th_blink_18;
        end
        th_blink_18: begin
          $display("result fifo.out = %d", _th_blink_read_data_33);
          th_blink <= th_blink_19;
        end
        th_blink_19: begin
          if(!myfifo_empty) begin
            th_blink <= th_blink_20;
          end 
        end
        th_blink_20: begin
          th_blink <= th_blink_21;
        end
        th_blink_21: begin
          if(_tmp_34) begin
            _tmp_35 <= myfifo_rdata;
          end 
          if(_tmp_34) begin
            th_blink <= th_blink_22;
          end 
        end
        th_blink_22: begin
          _th_blink_read_data_33 <= _tmp_35;
          th_blink <= th_blink_23;
        end
        th_blink_23: begin
          LED <= _th_blink_read_data_33;
          th_blink <= th_blink_24;
        end
        th_blink_24: begin
          $display("result fifo.out = %d", _th_blink_read_data_33);
          th_blink <= th_blink_25;
        end
      endcase
    end
  end

  localparam th_myfunc_0_1 = 1;
  localparam th_myfunc_0_2 = 2;
  localparam th_myfunc_0_3 = 3;
  localparam th_myfunc_0_4 = 4;
  localparam th_myfunc_0_5 = 5;
  localparam th_myfunc_0_6 = 6;
  localparam th_myfunc_0_7 = 7;
  localparam th_myfunc_0_8 = 8;
  localparam th_myfunc_0_9 = 9;
  localparam th_myfunc_0_10 = 10;
  localparam th_myfunc_0_11 = 11;
  localparam th_myfunc_0_12 = 12;
  localparam th_myfunc_0_13 = 13;
  localparam th_myfunc_0_14 = 14;
  localparam th_myfunc_0_15 = 15;
  localparam th_myfunc_0_16 = 16;
  localparam th_myfunc_0_17 = 17;
  localparam th_myfunc_0_18 = 18;
  localparam th_myfunc_0_19 = 19;
  localparam th_myfunc_0_20 = 20;
  localparam th_myfunc_0_21 = 21;
  localparam th_myfunc_0_22 = 22;
  localparam th_myfunc_0_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_0 <= th_myfunc_0_init;
      _th_myfunc_0_called <= 0;
      _th_myfunc_0_tid_1 <= 0;
      _th_myfunc_0_tid_2 <= 0;
      _tmp_1 <= 0;
      _th_myfunc_0_read_data_3 <= 0;
      _th_myfunc_0_write_data_4 <= 0;
      _tmp_3 <= 0;
    end else begin
      case(th_myfunc_0)
        th_myfunc_0_init: begin
          if(_th_myfunc_start[0] && (th_blink == 6)) begin
            _th_myfunc_0_called <= 1;
          end 
          if(_th_myfunc_start[0] && (th_blink == 6)) begin
            _th_myfunc_0_tid_1 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[0]) begin
            th_myfunc_0 <= th_myfunc_0_1;
          end 
        end
        th_myfunc_0_1: begin
          _th_myfunc_0_tid_2 <= _th_myfunc_0_tid_1;
          th_myfunc_0 <= th_myfunc_0_2;
        end
        th_myfunc_0_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 0)) begin
            th_myfunc_0 <= th_myfunc_0_3;
          end 
        end
        th_myfunc_0_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 0))) begin
            th_myfunc_0 <= th_myfunc_0_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 0)) begin
            th_myfunc_0 <= th_myfunc_0_4;
          end 
        end
        th_myfunc_0_4: begin
          $display("Thread %d Lock", _th_myfunc_0_tid_2);
          th_myfunc_0 <= th_myfunc_0_5;
        end
        th_myfunc_0_5: begin
          if(!myfifo_empty) begin
            th_myfunc_0 <= th_myfunc_0_6;
          end 
        end
        th_myfunc_0_6: begin
          th_myfunc_0 <= th_myfunc_0_7;
        end
        th_myfunc_0_7: begin
          if(_tmp_0) begin
            _tmp_1 <= myfifo_rdata;
          end 
          if(_tmp_0) begin
            th_myfunc_0 <= th_myfunc_0_8;
          end 
        end
        th_myfunc_0_8: begin
          _th_myfunc_0_read_data_3 <= _tmp_1;
          th_myfunc_0 <= th_myfunc_0_9;
        end
        th_myfunc_0_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_0_tid_2, _th_myfunc_0_read_data_3);
          th_myfunc_0 <= th_myfunc_0_10;
        end
        th_myfunc_0_10: begin
          _th_myfunc_0_write_data_4 <= _th_myfunc_0_read_data_3 + 1;
          th_myfunc_0 <= th_myfunc_0_11;
        end
        th_myfunc_0_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_0 <= th_myfunc_0_12;
          end 
        end
        th_myfunc_0_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_0_tid_2, _th_myfunc_0_write_data_4);
          th_myfunc_0 <= th_myfunc_0_13;
        end
        th_myfunc_0_13: begin
          if(!myfifo_empty) begin
            th_myfunc_0 <= th_myfunc_0_14;
          end 
        end
        th_myfunc_0_14: begin
          th_myfunc_0 <= th_myfunc_0_15;
        end
        th_myfunc_0_15: begin
          if(_tmp_2) begin
            _tmp_3 <= myfifo_rdata;
          end 
          if(_tmp_2) begin
            th_myfunc_0 <= th_myfunc_0_16;
          end 
        end
        th_myfunc_0_16: begin
          _th_myfunc_0_read_data_3 <= _tmp_3;
          th_myfunc_0 <= th_myfunc_0_17;
        end
        th_myfunc_0_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_0_tid_2, _th_myfunc_0_read_data_3);
          th_myfunc_0 <= th_myfunc_0_18;
        end
        th_myfunc_0_18: begin
          _th_myfunc_0_write_data_4 <= _th_myfunc_0_read_data_3 + 1;
          th_myfunc_0 <= th_myfunc_0_19;
        end
        th_myfunc_0_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_0 <= th_myfunc_0_20;
          end 
        end
        th_myfunc_0_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_0_tid_2, _th_myfunc_0_write_data_4);
          th_myfunc_0 <= th_myfunc_0_21;
        end
        th_myfunc_0_21: begin
          th_myfunc_0 <= th_myfunc_0_22;
        end
        th_myfunc_0_22: begin
          $display("Thread %d Unlock", _th_myfunc_0_tid_2);
          th_myfunc_0 <= th_myfunc_0_23;
        end
      endcase
    end
  end

  localparam th_myfunc_1_1 = 1;
  localparam th_myfunc_1_2 = 2;
  localparam th_myfunc_1_3 = 3;
  localparam th_myfunc_1_4 = 4;
  localparam th_myfunc_1_5 = 5;
  localparam th_myfunc_1_6 = 6;
  localparam th_myfunc_1_7 = 7;
  localparam th_myfunc_1_8 = 8;
  localparam th_myfunc_1_9 = 9;
  localparam th_myfunc_1_10 = 10;
  localparam th_myfunc_1_11 = 11;
  localparam th_myfunc_1_12 = 12;
  localparam th_myfunc_1_13 = 13;
  localparam th_myfunc_1_14 = 14;
  localparam th_myfunc_1_15 = 15;
  localparam th_myfunc_1_16 = 16;
  localparam th_myfunc_1_17 = 17;
  localparam th_myfunc_1_18 = 18;
  localparam th_myfunc_1_19 = 19;
  localparam th_myfunc_1_20 = 20;
  localparam th_myfunc_1_21 = 21;
  localparam th_myfunc_1_22 = 22;
  localparam th_myfunc_1_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_1 <= th_myfunc_1_init;
      _th_myfunc_1_called <= 0;
      _th_myfunc_1_tid_5 <= 0;
      _th_myfunc_1_tid_6 <= 0;
      _tmp_5 <= 0;
      _th_myfunc_1_read_data_7 <= 0;
      _th_myfunc_1_write_data_8 <= 0;
      _tmp_7 <= 0;
    end else begin
      case(th_myfunc_1)
        th_myfunc_1_init: begin
          if(_th_myfunc_start[1] && (th_blink == 6)) begin
            _th_myfunc_1_called <= 1;
          end 
          if(_th_myfunc_start[1] && (th_blink == 6)) begin
            _th_myfunc_1_tid_5 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[1]) begin
            th_myfunc_1 <= th_myfunc_1_1;
          end 
        end
        th_myfunc_1_1: begin
          _th_myfunc_1_tid_6 <= _th_myfunc_1_tid_5;
          th_myfunc_1 <= th_myfunc_1_2;
        end
        th_myfunc_1_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 1)) begin
            th_myfunc_1 <= th_myfunc_1_3;
          end 
        end
        th_myfunc_1_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 1))) begin
            th_myfunc_1 <= th_myfunc_1_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 1)) begin
            th_myfunc_1 <= th_myfunc_1_4;
          end 
        end
        th_myfunc_1_4: begin
          $display("Thread %d Lock", _th_myfunc_1_tid_6);
          th_myfunc_1 <= th_myfunc_1_5;
        end
        th_myfunc_1_5: begin
          if(!myfifo_empty) begin
            th_myfunc_1 <= th_myfunc_1_6;
          end 
        end
        th_myfunc_1_6: begin
          th_myfunc_1 <= th_myfunc_1_7;
        end
        th_myfunc_1_7: begin
          if(_tmp_4) begin
            _tmp_5 <= myfifo_rdata;
          end 
          if(_tmp_4) begin
            th_myfunc_1 <= th_myfunc_1_8;
          end 
        end
        th_myfunc_1_8: begin
          _th_myfunc_1_read_data_7 <= _tmp_5;
          th_myfunc_1 <= th_myfunc_1_9;
        end
        th_myfunc_1_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_1_tid_6, _th_myfunc_1_read_data_7);
          th_myfunc_1 <= th_myfunc_1_10;
        end
        th_myfunc_1_10: begin
          _th_myfunc_1_write_data_8 <= _th_myfunc_1_read_data_7 + 1;
          th_myfunc_1 <= th_myfunc_1_11;
        end
        th_myfunc_1_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_1 <= th_myfunc_1_12;
          end 
        end
        th_myfunc_1_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_1_tid_6, _th_myfunc_1_write_data_8);
          th_myfunc_1 <= th_myfunc_1_13;
        end
        th_myfunc_1_13: begin
          if(!myfifo_empty) begin
            th_myfunc_1 <= th_myfunc_1_14;
          end 
        end
        th_myfunc_1_14: begin
          th_myfunc_1 <= th_myfunc_1_15;
        end
        th_myfunc_1_15: begin
          if(_tmp_6) begin
            _tmp_7 <= myfifo_rdata;
          end 
          if(_tmp_6) begin
            th_myfunc_1 <= th_myfunc_1_16;
          end 
        end
        th_myfunc_1_16: begin
          _th_myfunc_1_read_data_7 <= _tmp_7;
          th_myfunc_1 <= th_myfunc_1_17;
        end
        th_myfunc_1_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_1_tid_6, _th_myfunc_1_read_data_7);
          th_myfunc_1 <= th_myfunc_1_18;
        end
        th_myfunc_1_18: begin
          _th_myfunc_1_write_data_8 <= _th_myfunc_1_read_data_7 + 1;
          th_myfunc_1 <= th_myfunc_1_19;
        end
        th_myfunc_1_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_1 <= th_myfunc_1_20;
          end 
        end
        th_myfunc_1_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_1_tid_6, _th_myfunc_1_write_data_8);
          th_myfunc_1 <= th_myfunc_1_21;
        end
        th_myfunc_1_21: begin
          th_myfunc_1 <= th_myfunc_1_22;
        end
        th_myfunc_1_22: begin
          $display("Thread %d Unlock", _th_myfunc_1_tid_6);
          th_myfunc_1 <= th_myfunc_1_23;
        end
      endcase
    end
  end

  localparam th_myfunc_2_1 = 1;
  localparam th_myfunc_2_2 = 2;
  localparam th_myfunc_2_3 = 3;
  localparam th_myfunc_2_4 = 4;
  localparam th_myfunc_2_5 = 5;
  localparam th_myfunc_2_6 = 6;
  localparam th_myfunc_2_7 = 7;
  localparam th_myfunc_2_8 = 8;
  localparam th_myfunc_2_9 = 9;
  localparam th_myfunc_2_10 = 10;
  localparam th_myfunc_2_11 = 11;
  localparam th_myfunc_2_12 = 12;
  localparam th_myfunc_2_13 = 13;
  localparam th_myfunc_2_14 = 14;
  localparam th_myfunc_2_15 = 15;
  localparam th_myfunc_2_16 = 16;
  localparam th_myfunc_2_17 = 17;
  localparam th_myfunc_2_18 = 18;
  localparam th_myfunc_2_19 = 19;
  localparam th_myfunc_2_20 = 20;
  localparam th_myfunc_2_21 = 21;
  localparam th_myfunc_2_22 = 22;
  localparam th_myfunc_2_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_2 <= th_myfunc_2_init;
      _th_myfunc_2_called <= 0;
      _th_myfunc_2_tid_9 <= 0;
      _th_myfunc_2_tid_10 <= 0;
      _tmp_9 <= 0;
      _th_myfunc_2_read_data_11 <= 0;
      _th_myfunc_2_write_data_12 <= 0;
      _tmp_11 <= 0;
    end else begin
      case(th_myfunc_2)
        th_myfunc_2_init: begin
          if(_th_myfunc_start[2] && (th_blink == 6)) begin
            _th_myfunc_2_called <= 1;
          end 
          if(_th_myfunc_start[2] && (th_blink == 6)) begin
            _th_myfunc_2_tid_9 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[2]) begin
            th_myfunc_2 <= th_myfunc_2_1;
          end 
        end
        th_myfunc_2_1: begin
          _th_myfunc_2_tid_10 <= _th_myfunc_2_tid_9;
          th_myfunc_2 <= th_myfunc_2_2;
        end
        th_myfunc_2_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 2)) begin
            th_myfunc_2 <= th_myfunc_2_3;
          end 
        end
        th_myfunc_2_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 2))) begin
            th_myfunc_2 <= th_myfunc_2_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 2)) begin
            th_myfunc_2 <= th_myfunc_2_4;
          end 
        end
        th_myfunc_2_4: begin
          $display("Thread %d Lock", _th_myfunc_2_tid_10);
          th_myfunc_2 <= th_myfunc_2_5;
        end
        th_myfunc_2_5: begin
          if(!myfifo_empty) begin
            th_myfunc_2 <= th_myfunc_2_6;
          end 
        end
        th_myfunc_2_6: begin
          th_myfunc_2 <= th_myfunc_2_7;
        end
        th_myfunc_2_7: begin
          if(_tmp_8) begin
            _tmp_9 <= myfifo_rdata;
          end 
          if(_tmp_8) begin
            th_myfunc_2 <= th_myfunc_2_8;
          end 
        end
        th_myfunc_2_8: begin
          _th_myfunc_2_read_data_11 <= _tmp_9;
          th_myfunc_2 <= th_myfunc_2_9;
        end
        th_myfunc_2_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_2_tid_10, _th_myfunc_2_read_data_11);
          th_myfunc_2 <= th_myfunc_2_10;
        end
        th_myfunc_2_10: begin
          _th_myfunc_2_write_data_12 <= _th_myfunc_2_read_data_11 + 1;
          th_myfunc_2 <= th_myfunc_2_11;
        end
        th_myfunc_2_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_2 <= th_myfunc_2_12;
          end 
        end
        th_myfunc_2_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_2_tid_10, _th_myfunc_2_write_data_12);
          th_myfunc_2 <= th_myfunc_2_13;
        end
        th_myfunc_2_13: begin
          if(!myfifo_empty) begin
            th_myfunc_2 <= th_myfunc_2_14;
          end 
        end
        th_myfunc_2_14: begin
          th_myfunc_2 <= th_myfunc_2_15;
        end
        th_myfunc_2_15: begin
          if(_tmp_10) begin
            _tmp_11 <= myfifo_rdata;
          end 
          if(_tmp_10) begin
            th_myfunc_2 <= th_myfunc_2_16;
          end 
        end
        th_myfunc_2_16: begin
          _th_myfunc_2_read_data_11 <= _tmp_11;
          th_myfunc_2 <= th_myfunc_2_17;
        end
        th_myfunc_2_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_2_tid_10, _th_myfunc_2_read_data_11);
          th_myfunc_2 <= th_myfunc_2_18;
        end
        th_myfunc_2_18: begin
          _th_myfunc_2_write_data_12 <= _th_myfunc_2_read_data_11 + 1;
          th_myfunc_2 <= th_myfunc_2_19;
        end
        th_myfunc_2_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_2 <= th_myfunc_2_20;
          end 
        end
        th_myfunc_2_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_2_tid_10, _th_myfunc_2_write_data_12);
          th_myfunc_2 <= th_myfunc_2_21;
        end
        th_myfunc_2_21: begin
          th_myfunc_2 <= th_myfunc_2_22;
        end
        th_myfunc_2_22: begin
          $display("Thread %d Unlock", _th_myfunc_2_tid_10);
          th_myfunc_2 <= th_myfunc_2_23;
        end
      endcase
    end
  end

  localparam th_myfunc_3_1 = 1;
  localparam th_myfunc_3_2 = 2;
  localparam th_myfunc_3_3 = 3;
  localparam th_myfunc_3_4 = 4;
  localparam th_myfunc_3_5 = 5;
  localparam th_myfunc_3_6 = 6;
  localparam th_myfunc_3_7 = 7;
  localparam th_myfunc_3_8 = 8;
  localparam th_myfunc_3_9 = 9;
  localparam th_myfunc_3_10 = 10;
  localparam th_myfunc_3_11 = 11;
  localparam th_myfunc_3_12 = 12;
  localparam th_myfunc_3_13 = 13;
  localparam th_myfunc_3_14 = 14;
  localparam th_myfunc_3_15 = 15;
  localparam th_myfunc_3_16 = 16;
  localparam th_myfunc_3_17 = 17;
  localparam th_myfunc_3_18 = 18;
  localparam th_myfunc_3_19 = 19;
  localparam th_myfunc_3_20 = 20;
  localparam th_myfunc_3_21 = 21;
  localparam th_myfunc_3_22 = 22;
  localparam th_myfunc_3_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_3 <= th_myfunc_3_init;
      _th_myfunc_3_called <= 0;
      _th_myfunc_3_tid_13 <= 0;
      _th_myfunc_3_tid_14 <= 0;
      _tmp_13 <= 0;
      _th_myfunc_3_read_data_15 <= 0;
      _th_myfunc_3_write_data_16 <= 0;
      _tmp_15 <= 0;
    end else begin
      case(th_myfunc_3)
        th_myfunc_3_init: begin
          if(_th_myfunc_start[3] && (th_blink == 6)) begin
            _th_myfunc_3_called <= 1;
          end 
          if(_th_myfunc_start[3] && (th_blink == 6)) begin
            _th_myfunc_3_tid_13 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[3]) begin
            th_myfunc_3 <= th_myfunc_3_1;
          end 
        end
        th_myfunc_3_1: begin
          _th_myfunc_3_tid_14 <= _th_myfunc_3_tid_13;
          th_myfunc_3 <= th_myfunc_3_2;
        end
        th_myfunc_3_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 3)) begin
            th_myfunc_3 <= th_myfunc_3_3;
          end 
        end
        th_myfunc_3_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 3))) begin
            th_myfunc_3 <= th_myfunc_3_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 3)) begin
            th_myfunc_3 <= th_myfunc_3_4;
          end 
        end
        th_myfunc_3_4: begin
          $display("Thread %d Lock", _th_myfunc_3_tid_14);
          th_myfunc_3 <= th_myfunc_3_5;
        end
        th_myfunc_3_5: begin
          if(!myfifo_empty) begin
            th_myfunc_3 <= th_myfunc_3_6;
          end 
        end
        th_myfunc_3_6: begin
          th_myfunc_3 <= th_myfunc_3_7;
        end
        th_myfunc_3_7: begin
          if(_tmp_12) begin
            _tmp_13 <= myfifo_rdata;
          end 
          if(_tmp_12) begin
            th_myfunc_3 <= th_myfunc_3_8;
          end 
        end
        th_myfunc_3_8: begin
          _th_myfunc_3_read_data_15 <= _tmp_13;
          th_myfunc_3 <= th_myfunc_3_9;
        end
        th_myfunc_3_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_3_tid_14, _th_myfunc_3_read_data_15);
          th_myfunc_3 <= th_myfunc_3_10;
        end
        th_myfunc_3_10: begin
          _th_myfunc_3_write_data_16 <= _th_myfunc_3_read_data_15 + 1;
          th_myfunc_3 <= th_myfunc_3_11;
        end
        th_myfunc_3_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_3 <= th_myfunc_3_12;
          end 
        end
        th_myfunc_3_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_3_tid_14, _th_myfunc_3_write_data_16);
          th_myfunc_3 <= th_myfunc_3_13;
        end
        th_myfunc_3_13: begin
          if(!myfifo_empty) begin
            th_myfunc_3 <= th_myfunc_3_14;
          end 
        end
        th_myfunc_3_14: begin
          th_myfunc_3 <= th_myfunc_3_15;
        end
        th_myfunc_3_15: begin
          if(_tmp_14) begin
            _tmp_15 <= myfifo_rdata;
          end 
          if(_tmp_14) begin
            th_myfunc_3 <= th_myfunc_3_16;
          end 
        end
        th_myfunc_3_16: begin
          _th_myfunc_3_read_data_15 <= _tmp_15;
          th_myfunc_3 <= th_myfunc_3_17;
        end
        th_myfunc_3_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_3_tid_14, _th_myfunc_3_read_data_15);
          th_myfunc_3 <= th_myfunc_3_18;
        end
        th_myfunc_3_18: begin
          _th_myfunc_3_write_data_16 <= _th_myfunc_3_read_data_15 + 1;
          th_myfunc_3 <= th_myfunc_3_19;
        end
        th_myfunc_3_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_3 <= th_myfunc_3_20;
          end 
        end
        th_myfunc_3_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_3_tid_14, _th_myfunc_3_write_data_16);
          th_myfunc_3 <= th_myfunc_3_21;
        end
        th_myfunc_3_21: begin
          th_myfunc_3 <= th_myfunc_3_22;
        end
        th_myfunc_3_22: begin
          $display("Thread %d Unlock", _th_myfunc_3_tid_14);
          th_myfunc_3 <= th_myfunc_3_23;
        end
      endcase
    end
  end

  localparam th_myfunc_4_1 = 1;
  localparam th_myfunc_4_2 = 2;
  localparam th_myfunc_4_3 = 3;
  localparam th_myfunc_4_4 = 4;
  localparam th_myfunc_4_5 = 5;
  localparam th_myfunc_4_6 = 6;
  localparam th_myfunc_4_7 = 7;
  localparam th_myfunc_4_8 = 8;
  localparam th_myfunc_4_9 = 9;
  localparam th_myfunc_4_10 = 10;
  localparam th_myfunc_4_11 = 11;
  localparam th_myfunc_4_12 = 12;
  localparam th_myfunc_4_13 = 13;
  localparam th_myfunc_4_14 = 14;
  localparam th_myfunc_4_15 = 15;
  localparam th_myfunc_4_16 = 16;
  localparam th_myfunc_4_17 = 17;
  localparam th_myfunc_4_18 = 18;
  localparam th_myfunc_4_19 = 19;
  localparam th_myfunc_4_20 = 20;
  localparam th_myfunc_4_21 = 21;
  localparam th_myfunc_4_22 = 22;
  localparam th_myfunc_4_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_4 <= th_myfunc_4_init;
      _th_myfunc_4_called <= 0;
      _th_myfunc_4_tid_17 <= 0;
      _th_myfunc_4_tid_18 <= 0;
      _tmp_17 <= 0;
      _th_myfunc_4_read_data_19 <= 0;
      _th_myfunc_4_write_data_20 <= 0;
      _tmp_19 <= 0;
    end else begin
      case(th_myfunc_4)
        th_myfunc_4_init: begin
          if(_th_myfunc_start[4] && (th_blink == 6)) begin
            _th_myfunc_4_called <= 1;
          end 
          if(_th_myfunc_start[4] && (th_blink == 6)) begin
            _th_myfunc_4_tid_17 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[4]) begin
            th_myfunc_4 <= th_myfunc_4_1;
          end 
        end
        th_myfunc_4_1: begin
          _th_myfunc_4_tid_18 <= _th_myfunc_4_tid_17;
          th_myfunc_4 <= th_myfunc_4_2;
        end
        th_myfunc_4_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 4)) begin
            th_myfunc_4 <= th_myfunc_4_3;
          end 
        end
        th_myfunc_4_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 4))) begin
            th_myfunc_4 <= th_myfunc_4_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 4)) begin
            th_myfunc_4 <= th_myfunc_4_4;
          end 
        end
        th_myfunc_4_4: begin
          $display("Thread %d Lock", _th_myfunc_4_tid_18);
          th_myfunc_4 <= th_myfunc_4_5;
        end
        th_myfunc_4_5: begin
          if(!myfifo_empty) begin
            th_myfunc_4 <= th_myfunc_4_6;
          end 
        end
        th_myfunc_4_6: begin
          th_myfunc_4 <= th_myfunc_4_7;
        end
        th_myfunc_4_7: begin
          if(_tmp_16) begin
            _tmp_17 <= myfifo_rdata;
          end 
          if(_tmp_16) begin
            th_myfunc_4 <= th_myfunc_4_8;
          end 
        end
        th_myfunc_4_8: begin
          _th_myfunc_4_read_data_19 <= _tmp_17;
          th_myfunc_4 <= th_myfunc_4_9;
        end
        th_myfunc_4_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_4_tid_18, _th_myfunc_4_read_data_19);
          th_myfunc_4 <= th_myfunc_4_10;
        end
        th_myfunc_4_10: begin
          _th_myfunc_4_write_data_20 <= _th_myfunc_4_read_data_19 + 1;
          th_myfunc_4 <= th_myfunc_4_11;
        end
        th_myfunc_4_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_4 <= th_myfunc_4_12;
          end 
        end
        th_myfunc_4_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_4_tid_18, _th_myfunc_4_write_data_20);
          th_myfunc_4 <= th_myfunc_4_13;
        end
        th_myfunc_4_13: begin
          if(!myfifo_empty) begin
            th_myfunc_4 <= th_myfunc_4_14;
          end 
        end
        th_myfunc_4_14: begin
          th_myfunc_4 <= th_myfunc_4_15;
        end
        th_myfunc_4_15: begin
          if(_tmp_18) begin
            _tmp_19 <= myfifo_rdata;
          end 
          if(_tmp_18) begin
            th_myfunc_4 <= th_myfunc_4_16;
          end 
        end
        th_myfunc_4_16: begin
          _th_myfunc_4_read_data_19 <= _tmp_19;
          th_myfunc_4 <= th_myfunc_4_17;
        end
        th_myfunc_4_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_4_tid_18, _th_myfunc_4_read_data_19);
          th_myfunc_4 <= th_myfunc_4_18;
        end
        th_myfunc_4_18: begin
          _th_myfunc_4_write_data_20 <= _th_myfunc_4_read_data_19 + 1;
          th_myfunc_4 <= th_myfunc_4_19;
        end
        th_myfunc_4_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_4 <= th_myfunc_4_20;
          end 
        end
        th_myfunc_4_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_4_tid_18, _th_myfunc_4_write_data_20);
          th_myfunc_4 <= th_myfunc_4_21;
        end
        th_myfunc_4_21: begin
          th_myfunc_4 <= th_myfunc_4_22;
        end
        th_myfunc_4_22: begin
          $display("Thread %d Unlock", _th_myfunc_4_tid_18);
          th_myfunc_4 <= th_myfunc_4_23;
        end
      endcase
    end
  end

  localparam th_myfunc_5_1 = 1;
  localparam th_myfunc_5_2 = 2;
  localparam th_myfunc_5_3 = 3;
  localparam th_myfunc_5_4 = 4;
  localparam th_myfunc_5_5 = 5;
  localparam th_myfunc_5_6 = 6;
  localparam th_myfunc_5_7 = 7;
  localparam th_myfunc_5_8 = 8;
  localparam th_myfunc_5_9 = 9;
  localparam th_myfunc_5_10 = 10;
  localparam th_myfunc_5_11 = 11;
  localparam th_myfunc_5_12 = 12;
  localparam th_myfunc_5_13 = 13;
  localparam th_myfunc_5_14 = 14;
  localparam th_myfunc_5_15 = 15;
  localparam th_myfunc_5_16 = 16;
  localparam th_myfunc_5_17 = 17;
  localparam th_myfunc_5_18 = 18;
  localparam th_myfunc_5_19 = 19;
  localparam th_myfunc_5_20 = 20;
  localparam th_myfunc_5_21 = 21;
  localparam th_myfunc_5_22 = 22;
  localparam th_myfunc_5_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_5 <= th_myfunc_5_init;
      _th_myfunc_5_called <= 0;
      _th_myfunc_5_tid_21 <= 0;
      _th_myfunc_5_tid_22 <= 0;
      _tmp_21 <= 0;
      _th_myfunc_5_read_data_23 <= 0;
      _th_myfunc_5_write_data_24 <= 0;
      _tmp_23 <= 0;
    end else begin
      case(th_myfunc_5)
        th_myfunc_5_init: begin
          if(_th_myfunc_start[5] && (th_blink == 6)) begin
            _th_myfunc_5_called <= 1;
          end 
          if(_th_myfunc_start[5] && (th_blink == 6)) begin
            _th_myfunc_5_tid_21 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[5]) begin
            th_myfunc_5 <= th_myfunc_5_1;
          end 
        end
        th_myfunc_5_1: begin
          _th_myfunc_5_tid_22 <= _th_myfunc_5_tid_21;
          th_myfunc_5 <= th_myfunc_5_2;
        end
        th_myfunc_5_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 5)) begin
            th_myfunc_5 <= th_myfunc_5_3;
          end 
        end
        th_myfunc_5_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 5))) begin
            th_myfunc_5 <= th_myfunc_5_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 5)) begin
            th_myfunc_5 <= th_myfunc_5_4;
          end 
        end
        th_myfunc_5_4: begin
          $display("Thread %d Lock", _th_myfunc_5_tid_22);
          th_myfunc_5 <= th_myfunc_5_5;
        end
        th_myfunc_5_5: begin
          if(!myfifo_empty) begin
            th_myfunc_5 <= th_myfunc_5_6;
          end 
        end
        th_myfunc_5_6: begin
          th_myfunc_5 <= th_myfunc_5_7;
        end
        th_myfunc_5_7: begin
          if(_tmp_20) begin
            _tmp_21 <= myfifo_rdata;
          end 
          if(_tmp_20) begin
            th_myfunc_5 <= th_myfunc_5_8;
          end 
        end
        th_myfunc_5_8: begin
          _th_myfunc_5_read_data_23 <= _tmp_21;
          th_myfunc_5 <= th_myfunc_5_9;
        end
        th_myfunc_5_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_5_tid_22, _th_myfunc_5_read_data_23);
          th_myfunc_5 <= th_myfunc_5_10;
        end
        th_myfunc_5_10: begin
          _th_myfunc_5_write_data_24 <= _th_myfunc_5_read_data_23 + 1;
          th_myfunc_5 <= th_myfunc_5_11;
        end
        th_myfunc_5_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_5 <= th_myfunc_5_12;
          end 
        end
        th_myfunc_5_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_5_tid_22, _th_myfunc_5_write_data_24);
          th_myfunc_5 <= th_myfunc_5_13;
        end
        th_myfunc_5_13: begin
          if(!myfifo_empty) begin
            th_myfunc_5 <= th_myfunc_5_14;
          end 
        end
        th_myfunc_5_14: begin
          th_myfunc_5 <= th_myfunc_5_15;
        end
        th_myfunc_5_15: begin
          if(_tmp_22) begin
            _tmp_23 <= myfifo_rdata;
          end 
          if(_tmp_22) begin
            th_myfunc_5 <= th_myfunc_5_16;
          end 
        end
        th_myfunc_5_16: begin
          _th_myfunc_5_read_data_23 <= _tmp_23;
          th_myfunc_5 <= th_myfunc_5_17;
        end
        th_myfunc_5_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_5_tid_22, _th_myfunc_5_read_data_23);
          th_myfunc_5 <= th_myfunc_5_18;
        end
        th_myfunc_5_18: begin
          _th_myfunc_5_write_data_24 <= _th_myfunc_5_read_data_23 + 1;
          th_myfunc_5 <= th_myfunc_5_19;
        end
        th_myfunc_5_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_5 <= th_myfunc_5_20;
          end 
        end
        th_myfunc_5_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_5_tid_22, _th_myfunc_5_write_data_24);
          th_myfunc_5 <= th_myfunc_5_21;
        end
        th_myfunc_5_21: begin
          th_myfunc_5 <= th_myfunc_5_22;
        end
        th_myfunc_5_22: begin
          $display("Thread %d Unlock", _th_myfunc_5_tid_22);
          th_myfunc_5 <= th_myfunc_5_23;
        end
      endcase
    end
  end

  localparam th_myfunc_6_1 = 1;
  localparam th_myfunc_6_2 = 2;
  localparam th_myfunc_6_3 = 3;
  localparam th_myfunc_6_4 = 4;
  localparam th_myfunc_6_5 = 5;
  localparam th_myfunc_6_6 = 6;
  localparam th_myfunc_6_7 = 7;
  localparam th_myfunc_6_8 = 8;
  localparam th_myfunc_6_9 = 9;
  localparam th_myfunc_6_10 = 10;
  localparam th_myfunc_6_11 = 11;
  localparam th_myfunc_6_12 = 12;
  localparam th_myfunc_6_13 = 13;
  localparam th_myfunc_6_14 = 14;
  localparam th_myfunc_6_15 = 15;
  localparam th_myfunc_6_16 = 16;
  localparam th_myfunc_6_17 = 17;
  localparam th_myfunc_6_18 = 18;
  localparam th_myfunc_6_19 = 19;
  localparam th_myfunc_6_20 = 20;
  localparam th_myfunc_6_21 = 21;
  localparam th_myfunc_6_22 = 22;
  localparam th_myfunc_6_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_6 <= th_myfunc_6_init;
      _th_myfunc_6_called <= 0;
      _th_myfunc_6_tid_25 <= 0;
      _th_myfunc_6_tid_26 <= 0;
      _tmp_25 <= 0;
      _th_myfunc_6_read_data_27 <= 0;
      _th_myfunc_6_write_data_28 <= 0;
      _tmp_27 <= 0;
    end else begin
      case(th_myfunc_6)
        th_myfunc_6_init: begin
          if(_th_myfunc_start[6] && (th_blink == 6)) begin
            _th_myfunc_6_called <= 1;
          end 
          if(_th_myfunc_start[6] && (th_blink == 6)) begin
            _th_myfunc_6_tid_25 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[6]) begin
            th_myfunc_6 <= th_myfunc_6_1;
          end 
        end
        th_myfunc_6_1: begin
          _th_myfunc_6_tid_26 <= _th_myfunc_6_tid_25;
          th_myfunc_6 <= th_myfunc_6_2;
        end
        th_myfunc_6_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 6)) begin
            th_myfunc_6 <= th_myfunc_6_3;
          end 
        end
        th_myfunc_6_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 6))) begin
            th_myfunc_6 <= th_myfunc_6_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 6)) begin
            th_myfunc_6 <= th_myfunc_6_4;
          end 
        end
        th_myfunc_6_4: begin
          $display("Thread %d Lock", _th_myfunc_6_tid_26);
          th_myfunc_6 <= th_myfunc_6_5;
        end
        th_myfunc_6_5: begin
          if(!myfifo_empty) begin
            th_myfunc_6 <= th_myfunc_6_6;
          end 
        end
        th_myfunc_6_6: begin
          th_myfunc_6 <= th_myfunc_6_7;
        end
        th_myfunc_6_7: begin
          if(_tmp_24) begin
            _tmp_25 <= myfifo_rdata;
          end 
          if(_tmp_24) begin
            th_myfunc_6 <= th_myfunc_6_8;
          end 
        end
        th_myfunc_6_8: begin
          _th_myfunc_6_read_data_27 <= _tmp_25;
          th_myfunc_6 <= th_myfunc_6_9;
        end
        th_myfunc_6_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_6_tid_26, _th_myfunc_6_read_data_27);
          th_myfunc_6 <= th_myfunc_6_10;
        end
        th_myfunc_6_10: begin
          _th_myfunc_6_write_data_28 <= _th_myfunc_6_read_data_27 + 1;
          th_myfunc_6 <= th_myfunc_6_11;
        end
        th_myfunc_6_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_6 <= th_myfunc_6_12;
          end 
        end
        th_myfunc_6_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_6_tid_26, _th_myfunc_6_write_data_28);
          th_myfunc_6 <= th_myfunc_6_13;
        end
        th_myfunc_6_13: begin
          if(!myfifo_empty) begin
            th_myfunc_6 <= th_myfunc_6_14;
          end 
        end
        th_myfunc_6_14: begin
          th_myfunc_6 <= th_myfunc_6_15;
        end
        th_myfunc_6_15: begin
          if(_tmp_26) begin
            _tmp_27 <= myfifo_rdata;
          end 
          if(_tmp_26) begin
            th_myfunc_6 <= th_myfunc_6_16;
          end 
        end
        th_myfunc_6_16: begin
          _th_myfunc_6_read_data_27 <= _tmp_27;
          th_myfunc_6 <= th_myfunc_6_17;
        end
        th_myfunc_6_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_6_tid_26, _th_myfunc_6_read_data_27);
          th_myfunc_6 <= th_myfunc_6_18;
        end
        th_myfunc_6_18: begin
          _th_myfunc_6_write_data_28 <= _th_myfunc_6_read_data_27 + 1;
          th_myfunc_6 <= th_myfunc_6_19;
        end
        th_myfunc_6_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_6 <= th_myfunc_6_20;
          end 
        end
        th_myfunc_6_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_6_tid_26, _th_myfunc_6_write_data_28);
          th_myfunc_6 <= th_myfunc_6_21;
        end
        th_myfunc_6_21: begin
          th_myfunc_6 <= th_myfunc_6_22;
        end
        th_myfunc_6_22: begin
          $display("Thread %d Unlock", _th_myfunc_6_tid_26);
          th_myfunc_6 <= th_myfunc_6_23;
        end
      endcase
    end
  end

  localparam th_myfunc_7_1 = 1;
  localparam th_myfunc_7_2 = 2;
  localparam th_myfunc_7_3 = 3;
  localparam th_myfunc_7_4 = 4;
  localparam th_myfunc_7_5 = 5;
  localparam th_myfunc_7_6 = 6;
  localparam th_myfunc_7_7 = 7;
  localparam th_myfunc_7_8 = 8;
  localparam th_myfunc_7_9 = 9;
  localparam th_myfunc_7_10 = 10;
  localparam th_myfunc_7_11 = 11;
  localparam th_myfunc_7_12 = 12;
  localparam th_myfunc_7_13 = 13;
  localparam th_myfunc_7_14 = 14;
  localparam th_myfunc_7_15 = 15;
  localparam th_myfunc_7_16 = 16;
  localparam th_myfunc_7_17 = 17;
  localparam th_myfunc_7_18 = 18;
  localparam th_myfunc_7_19 = 19;
  localparam th_myfunc_7_20 = 20;
  localparam th_myfunc_7_21 = 21;
  localparam th_myfunc_7_22 = 22;
  localparam th_myfunc_7_23 = 23;

  always @(posedge CLK) begin
    if(RST) begin
      th_myfunc_7 <= th_myfunc_7_init;
      _th_myfunc_7_called <= 0;
      _th_myfunc_7_tid_29 <= 0;
      _th_myfunc_7_tid_30 <= 0;
      _tmp_29 <= 0;
      _th_myfunc_7_read_data_31 <= 0;
      _th_myfunc_7_write_data_32 <= 0;
      _tmp_31 <= 0;
    end else begin
      case(th_myfunc_7)
        th_myfunc_7_init: begin
          if(_th_myfunc_start[7] && (th_blink == 6)) begin
            _th_myfunc_7_called <= 1;
          end 
          if(_th_myfunc_start[7] && (th_blink == 6)) begin
            _th_myfunc_7_tid_29 <= _th_blink_tid_0;
          end 
          if((th_blink == 6) && _th_myfunc_start[7]) begin
            th_myfunc_7 <= th_myfunc_7_1;
          end 
        end
        th_myfunc_7_1: begin
          _th_myfunc_7_tid_30 <= _th_myfunc_7_tid_29;
          th_myfunc_7 <= th_myfunc_7_2;
        end
        th_myfunc_7_2: begin
          if(!__myfifo_mutex_lock_reg || (__myfifo_mutex_lock_id == 7)) begin
            th_myfunc_7 <= th_myfunc_7_3;
          end 
        end
        th_myfunc_7_3: begin
          if(!(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 7))) begin
            th_myfunc_7 <= th_myfunc_7_2;
          end 
          if(__myfifo_mutex_lock_reg && (__myfifo_mutex_lock_id == 7)) begin
            th_myfunc_7 <= th_myfunc_7_4;
          end 
        end
        th_myfunc_7_4: begin
          $display("Thread %d Lock", _th_myfunc_7_tid_30);
          th_myfunc_7 <= th_myfunc_7_5;
        end
        th_myfunc_7_5: begin
          if(!myfifo_empty) begin
            th_myfunc_7 <= th_myfunc_7_6;
          end 
        end
        th_myfunc_7_6: begin
          th_myfunc_7 <= th_myfunc_7_7;
        end
        th_myfunc_7_7: begin
          if(_tmp_28) begin
            _tmp_29 <= myfifo_rdata;
          end 
          if(_tmp_28) begin
            th_myfunc_7 <= th_myfunc_7_8;
          end 
        end
        th_myfunc_7_8: begin
          _th_myfunc_7_read_data_31 <= _tmp_29;
          th_myfunc_7 <= th_myfunc_7_9;
        end
        th_myfunc_7_9: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_7_tid_30, _th_myfunc_7_read_data_31);
          th_myfunc_7 <= th_myfunc_7_10;
        end
        th_myfunc_7_10: begin
          _th_myfunc_7_write_data_32 <= _th_myfunc_7_read_data_31 + 1;
          th_myfunc_7 <= th_myfunc_7_11;
        end
        th_myfunc_7_11: begin
          if(!myfifo_almost_full) begin
            th_myfunc_7 <= th_myfunc_7_12;
          end 
        end
        th_myfunc_7_12: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_7_tid_30, _th_myfunc_7_write_data_32);
          th_myfunc_7 <= th_myfunc_7_13;
        end
        th_myfunc_7_13: begin
          if(!myfifo_empty) begin
            th_myfunc_7 <= th_myfunc_7_14;
          end 
        end
        th_myfunc_7_14: begin
          th_myfunc_7 <= th_myfunc_7_15;
        end
        th_myfunc_7_15: begin
          if(_tmp_30) begin
            _tmp_31 <= myfifo_rdata;
          end 
          if(_tmp_30) begin
            th_myfunc_7 <= th_myfunc_7_16;
          end 
        end
        th_myfunc_7_16: begin
          _th_myfunc_7_read_data_31 <= _tmp_31;
          th_myfunc_7 <= th_myfunc_7_17;
        end
        th_myfunc_7_17: begin
          $display("Thread %d fifo.out = %d", _th_myfunc_7_tid_30, _th_myfunc_7_read_data_31);
          th_myfunc_7 <= th_myfunc_7_18;
        end
        th_myfunc_7_18: begin
          _th_myfunc_7_write_data_32 <= _th_myfunc_7_read_data_31 + 1;
          th_myfunc_7 <= th_myfunc_7_19;
        end
        th_myfunc_7_19: begin
          if(!myfifo_almost_full) begin
            th_myfunc_7 <= th_myfunc_7_20;
          end 
        end
        th_myfunc_7_20: begin
          $display("Thread %d fifo.in <- %d", _th_myfunc_7_tid_30, _th_myfunc_7_write_data_32);
          th_myfunc_7 <= th_myfunc_7_21;
        end
        th_myfunc_7_21: begin
          th_myfunc_7 <= th_myfunc_7_22;
        end
        th_myfunc_7_22: begin
          $display("Thread %d Unlock", _th_myfunc_7_tid_30);
          th_myfunc_7 <= th_myfunc_7_23;
        end
      endcase
    end
  end


  always @(posedge CLK) begin
    if(RST) begin
      __myfifo_mutex_lock_reg <= 0;
      __myfifo_mutex_lock_id <= 0;
    end else begin
      if((th_myfunc_0 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 0;
      end 
      if((th_myfunc_0 == 21) && (__myfifo_mutex_lock_id == 0)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_1 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 1;
      end 
      if((th_myfunc_1 == 21) && (__myfifo_mutex_lock_id == 1)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_2 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 2;
      end 
      if((th_myfunc_2 == 21) && (__myfifo_mutex_lock_id == 2)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_3 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 3;
      end 
      if((th_myfunc_3 == 21) && (__myfifo_mutex_lock_id == 3)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_4 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 4;
      end 
      if((th_myfunc_4 == 21) && (__myfifo_mutex_lock_id == 4)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_5 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 5;
      end 
      if((th_myfunc_5 == 21) && (__myfifo_mutex_lock_id == 5)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_6 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 6;
      end 
      if((th_myfunc_6 == 21) && (__myfifo_mutex_lock_id == 6)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
      if((th_myfunc_7 == 2) && !__myfifo_mutex_lock_reg) begin
        __myfifo_mutex_lock_reg <= 1;
        __myfifo_mutex_lock_id <= 7;
      end 
      if((th_myfunc_7 == 21) && (__myfifo_mutex_lock_id == 7)) begin
        __myfifo_mutex_lock_reg <= 0;
      end 
    end
  end


endmodule



module myfifo
(
  input CLK,
  input RST,
  input myfifo_enq,
  input [32-1:0] myfifo_wdata,
  output myfifo_full,
  output myfifo_almost_full,
  input myfifo_deq,
  output [32-1:0] myfifo_rdata,
  output myfifo_empty,
  output myfifo_almost_empty
);

  reg [32-1:0] mem [0:16-1];
  reg [4-1:0] head;
  reg [4-1:0] tail;
  wire is_empty;
  wire is_almost_empty;
  wire is_full;
  wire is_almost_full;
  assign is_empty = head == tail;
  assign is_almost_empty = head == (tail + 1 & 15);
  assign is_full = (head + 1 & 15) == tail;
  assign is_almost_full = (head + 2 & 15) == tail;
  reg [32-1:0] rdata_reg;
  assign myfifo_full = is_full;
  assign myfifo_almost_full = is_almost_full || is_full;
  assign myfifo_empty = is_empty;
  assign myfifo_almost_empty = is_almost_empty || is_empty;
  assign myfifo_rdata = rdata_reg;

  always @(posedge CLK) begin
    if(RST) begin
      head <= 0;
      rdata_reg <= 0;
      tail <= 0;
    end else begin
      if(myfifo_enq && !is_full) begin
        mem[head] <= myfifo_wdata;
        head <= head + 1;
      end 
      if(myfifo_deq && !is_empty) begin
        rdata_reg <= mem[tail];
        tail <= tail + 1;
      end 
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = thread_fifo_multithread.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function

import copy
import veriloggen.core.vtypes as vtypes
from veriloggen.core.module import Module
from . import util


def mkRAMDefinition(name, datawidth=32, addrwidth=10, numports=2,
                    initvals=None, sync=True, with_enable=False,
                    nocheck_initvals=False):
    m = Module(name)
    clk = m.Input('CLK')

    interfaces = []

    for i in range(numports):
        interface = RAMSlaveInterface(
            m, name + '_%d' % i, datawidth, addrwidth, with_enable=with_enable)
        if sync:
            interface.delay_addr = m.Reg(name + '_%d_daddr' % i, addrwidth)

        interfaces.append(interface)

    mem = m.Reg('mem', datawidth, length=2**addrwidth)

    if initvals is not None:
        if not isinstance(initvals, (tuple, list)):
            raise TypeError("initvals must be tuple or list, not '%s" %
                            str(type(initvals)))

        base = 16

        if not nocheck_initvals:
            new_initvals = []
            for initval in initvals:
                if isinstance(initval, int):
                    new_initvals.append(
                        vtypes.Int(initval, datawidth, base=16))
                elif isinstance(initval, vtypes.Int) and isinstance(initval.value, int):
                    v = copy.deepcopy(initval)
                    v.width = datawidth
                    v.base = base
                    new_initvals.append(v)
                elif isinstance(initval, vtypes.Int) and isinstance(initval.value, str):
                    v = copy.deepcopy(initval)
                    v.width = datawidth
                    if v.base != 2 and v.base != 16:
                        raise ValueError('base must be 2 or 16')
                    base = v.base
                    new_initvals.append(v)
                else:
                    raise TypeError("values of initvals must be int, not '%s" %
                                    str(type(initval)))

            initvals = new_initvals

        if 2 ** addrwidth > len(initvals):
            initvals.extend(
                [vtypes.Int(0, datawidth, base=base)
                 for _ in range(2 ** addrwidth - len(initvals))])

        m.Initial(
            *[mem[i](initval) for i, initval in enumerate(initvals)]
        )

    for interface in interfaces:
        body = [
            vtypes.If(interface.wenable)(
                mem[interface.addr](interface.wdata)
            )]

        if sync:
            body.append(interface.delay_addr(interface.addr))

        if with_enable:
            body = vtypes.If(interface.enable)(*body)

        m.Always(vtypes.Posedge(clk))(
            body
        )

        if sync:
            m.Assign(interface.rdata(mem[interface.delay_addr]))
        else:
            m.Assign(interface.rdata(mem[interface.addr]))

    return m


class RAMInterface(object):
    _I = 'Reg'
    _O = 'Wire'

    def __init__(self, m, name=None, datawidth=32, addrwidth=10,
                 itype=None, otype=None,
                 p_addr='addr', p_rdata='rdata',
                 p_wdata='wdata', p_wenable='wenable',
                 p_enable='enable',
                 with_enable=False, index=None):

        if itype is None:
            itype = self._I
        if otype is None:
            otype = self._O

        self.m = m

        name_addr = p_addr if name is None else '_'.join([name, p_addr])
        name_rdata = p_rdata if name is None else '_'.join([name, p_rdata])
        name_wdata = p_wdata if name is None else '_'.join([name, p_wdata])
        name_wenable = (
            p_wenable if name is None else '_'.join([name, p_wenable]))
        if with_enable:
            name_enable = (
                p_enable if name is None else '_'.join([name, p_enable]))

        if index is not None:
            name_addr = name_addr + str(index)
            name_rdata = name_rdata + str(index)
            name_wdata = name_wdata + str(index)
            name_wenable = name_wenable + str(index)
            if with_enable:
                name_enable = name_enable + str(index)

        self.addr = util.make_port(m, itype, name_addr, addrwidth, initval=0)
        self.rdata = util.make_port(m, otype, name_rdata, datawidth, initval=0)
        self.wdata = util.make_port(m, itype, name_wdata, datawidth, initval=0)
        self.wenable = util.make_port(m, itype, name_wenable, initval=0)
        if with_enable:
            self.enable = util.make_port(m, itype, name_enable, initval=0)

    def connect(self, targ):
        self.addr.connect(targ.addr)
        targ.rdata.connect(self.rdata)
        self.wdata.connect(targ.wdata)
        self.wenable.connect(targ.wenable)
        if hasattr(self, 'enable'):
            if hasattr(targ, 'enable'):
                self.enable.connect(targ.enable)
            else:
                self.enable.connect(1)
        else:
            if hasattr(targ, 'enable'):
                raise ValueError('no enable port')


class RAMSlaveInterface(RAMInterface):
    _I = 'Input'
    _O = 'Output'


class RAMMasterInterface(RAMInterface):
    _I = 'Output'
    _O = 'Input'


class _RAM_RTL(object):

    def __init__(self, m, name, clk,
                 datawidth=32, addrwidth=10, numports=1,
                 initvals=None, sync=True, with_enable=False):

        self.m = m
        self.name = name
        self.clk = clk
        self.with_enable = with_enable

        self.interfaces = [RAMInterface(m, name + '_%d' % i, datawidth, addrwidth,
                                        itype='Wire', otype='Wire', with_enable=with_enable)
                           for i in range(numports)]

        ram_def = mkRAMDefinition(name, datawidth, addrwidth, numports,
                                  initvals, sync, with_enable)

        self.m.Instance(ram_def, name,
                        params=(), ports=m.connect_ports(ram_def))

    def connect(self, port, addr, wdata, wenable, enable=None):
        self.m.Assign(self.interfaces[port].addr(addr))
        self.m.Assign(self.interfaces[port].wdata(wdata))
        self.m.Assign(self.interfaces[port].wenable(wenable))
        if self.with_enable:
            self.m.Assign(self.interfaces[port].enable(enable))

    def rdata(self, port):
        return self.interfaces[port].rdata


class SyncRAM(_RAM_RTL):

    def __init__(self, m, name, clk,
                 datawidth=32, addrwidth=10, numports=1,
                 initvals=None, with_enable=False):
        _RAM_RTL.__init__(self, m, name, clk,
                          datawidth, addrwidth, numports,
                          initvals, sync=True, with_enable=with_enable)


class AsyncRAM(_RAM_RTL):

    def __init__(self, m, name, clk,
                 datawidth=32, addrwidth=10, numports=1,
                 initvals=None, with_enable=False):
        _RAM_RTL.__init__(self, m, name, clk,
                          datawidth, addrwidth, numports,
                          initvals, sync=False)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import fsm_goto_from

expected_verilog = """
module blinkled #
  (
   parameter WIDTH = 8
   )
  (
   input CLK, 
   input RST, 
   output reg [WIDTH-1:0] LED
   );
  reg [32-1:0] count;
  reg [32-1:0] fsm;
  localparam fsm_init = 0;
  localparam fsm_1 = 1;
  localparam fsm_2 = 2;
  localparam fsm_3 = 3;
  localparam fsm_4 = 4;
  localparam fsm_5 = 5;
  localparam fsm_6 = 6;
  localparam fsm_7 = 7;

  always @(posedge CLK) begin
    if(RST) begin
      fsm <= fsm_init;
      LED <= 0;
      count <= 0;
    end else begin
      case(fsm)
        fsm_init: begin
          LED <= 0;
          fsm <= fsm_1;
        end
        fsm_1: begin
          count <= 0;
          fsm <= fsm_2;
        end
        fsm_2: begin
          if(count < 1024) begin
            fsm <= fsm_3;
          end 
          if(!(count < 1024)) begin
            fsm <= fsm_7;
          end 
        end
        fsm_3: begin
          count <= count + 1;
          fsm <= fsm_4;
        end
        fsm_4: begin
          count <= count + 1;
          fsm <= fsm_5;
        end
        fsm_5: begin
          count <= count + 1;
          fsm <= fsm_6;
        end
        fsm_6: begin
          count <= count + 1;
          fsm <= fsm_2;
        end
        fsm_7: begin
          LED <= LED + 1;
          fsm <= fsm_1;
        end
      endcase
    end
  end

endmodule
"""

def test():
    veriloggen.reset()
    test_module = fsm_goto_from.mkLed()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import veriloggen
import stream_add_validready

expected_verilog = """
module test;

  reg CLK;
  reg RST;
  reg ivalid;
  wire iready;
  wire ovalid;
  reg oready;
  reg signed [32-1:0] xdata;
  reg signed [32-1:0] ydata;
  wire signed [32-1:0] zdata;

  main
  uut
  (
    .CLK(CLK),
    .RST(RST),
    .ivalid(ivalid),
    .iready(iready),
    .ovalid(ovalid),
    .oready(oready),
    .xdata(xdata),
    .ydata(ydata),
    .zdata(zdata)
  );

  reg reset_done;

  initial begin
    $dumpfile("uut.vcd");
    $dumpvars(0, uut);
  end


  initial begin
    CLK = 0;
    forever begin
      #5 CLK = !CLK;
    end
  end


  initial begin
    RST = 0;
    reset_done = 0;
    xdata = 0;
    ydata = 0;
    ivalid = 0;
    oready = 0;
    #100;
    RST = 1;
    #100;
    RST = 0;
    #1000;
    reset_done = 1;
    @(posedge CLK);
    #1;
    #10000;
    $finish;
  end

  reg [32-1:0] send_fsm;
  localparam send_fsm_init = 0;
  reg [32-1:0] send_count;
  reg [32-1:0] recv_fsm;
  localparam recv_fsm_init = 0;
  reg [32-1:0] recv_count;
  localparam send_fsm_1 = 1;
  localparam send_fsm_2 = 2;
  localparam send_fsm_3 = 3;
  localparam send_fsm_4 = 4;

  always @(posedge CLK) begin
    if(RST) begin
      send_fsm <= send_fsm_init;
      send_count <= 0;
    end else begin
      case(send_fsm)
        send_fsm_init: begin
          if(reset_done) begin
            send_fsm <= send_fsm_1;
          end 
        end
        send_fsm_1: begin
          ivalid <= 0;
          send_count <= send_count + 1;
          if(send_count == 10) begin
            send_count <= 0;
          end 
          if(send_count == 10) begin
            send_fsm <= send_fsm_2;
          end 
        end
        send_fsm_2: begin
          xdata <= 0;
          ydata <= 0;
          ivalid <= 1;
          send_count <= send_count + 1;
          send_fsm <= send_fsm_3;
        end
        send_fsm_3: begin
          if(iready) begin
            xdata <= xdata + 1;
            ydata <= ydata + 2;
            ivalid <= 1;
            $display("xdata=%d", xdata);
            $display("ydata=%d", ydata);
            send_count <= send_count + 1;
          end 
          if(iready && (send_count == 20)) begin
            ivalid <= 0;
          end 
          if(iready && (send_count == 20)) begin
            send_fsm <= send_fsm_4;
          end 
        end
      endcase
    end
  end

  localparam recv_fsm_1 = 1;
  localparam recv_fsm_2 = 2;

  always @(posedge CLK) begin
    if(RST) begin
      recv_fsm <= recv_fsm_init;
      recv_count <= 0;
    end else begin
      case(recv_fsm)
        recv_fsm_init: begin
          if(reset_done) begin
            recv_fsm <= recv_fsm_1;
          end 
        end
        recv_fsm_1: begin
          recv_count <= recv_count + 1;
          if(recv_count == 20) begin
            recv_count <= 0;
          end 
          if(recv_count == 20) begin
            recv_fsm <= recv_fsm_2;
          end 
        end
        recv_fsm_2: begin
          oready <= !oready;
          if(ovalid && oready) begin
            $display("zdata=%d", zdata);
            recv_count <= recv_count + 1;
          end 
        end
      endcase
    end
  end


endmodule



module main
(
  input CLK,
  input RST,
  input ivalid,
  output iready,
  output ovalid,
  input oready,
  input signed [32-1:0] xdata,
  input signed [32-1:0] ydata,
  output signed [32-1:0] zdata
);

  wire _tmp_0;
  assign _tmp_0 = !ovalid || oready;
  reg _ivalid_0;
  assign ovalid = _ivalid_0;
  assign iready = _tmp_0;
  reg signed [32-1:0] _plus_data_2;
  assign zdata = _plus_data_2;

  always @(posedge CLK) begin
    if(RST) begin
      _ivalid_0 <= 0;
      _plus_data_2 <= 0;
    end else begin
      if(_tmp_0) begin
        _ivalid_0 <= ivalid;
      end 
      if(_tmp_0) begin
        _plus_data_2 <= xdata + ydata;
      end 
    end
  end


endmodule
"""


def test():
    veriloggen.reset()
    test_module = stream_add_validready.mkTest()
    code = test_module.to_verilog()

    from pyverilog.vparser.parser import VerilogParser
    from pyverilog.ast_code_generator.codegen import ASTCodeGenerator
    parser = VerilogParser()
    expected_ast = parser.parse(expected_verilog)
    codegen = ASTCodeGenerator()
    expected_code = codegen.visit(expected_ast)

    assert(expected_code == code)



from __future__ import absolute_import
from __future__ import print_function
import sys
import os
import math

# the next line can be removed after installation
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.dirname(
    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))))

from veriloggen import *
from veriloggen.thread import RAM
import veriloggen.dataflow as dataflow


def mkMain(n=128, datawidth=32, numports=2):
    m = Module('main')

    clk = m.Input('CLK')
    rst = m.Input('RST')

    addrwidth = int(math.log(n, 2)) * 2
    myram = RAM(m, 'myram', clk, rst, datawidth, addrwidth)

    df = dataflow.DataflowManager(m, clk, rst)
    # df.enable_draw_graph()

    fsm = FSM(m, 'fsm', clk, rst)

    length = 8
    a = df.Counter()
    b = df.Counter(size=length)
    c = b == 0

    wport = 0
    waddr = 0
    wlen = 32
    done = myram.write_dataflow(wport, waddr, a, wlen, cond=fsm, when=c)

    fsm.goto_next()
    fsm.If(done).goto_next()

    seq = Seq(m, 'seq', clk, rst)

    seq.If(myram[0].wenable)(
        Systask('display', '[%d] <- %d', myram[0].addr, myram[0].wdata)
    )

    return m


def mkTest():
    m = Module('test')

    # target instance
    main = mkMain()

    # copy paras and ports
    params = m.copy_params(main)
    ports = m.copy_sim_ports(main)

    clk = ports['CLK']
    rst = ports['RST']

    uut = m.Instance(main, 'uut',
                     params=m.connect_params(main),
                     ports=m.connect_ports(main))

    simulation.setup_waveform(m, uut, m.get_vars())
    simulation.setup_clock(m, clk, hperiod=5)
    init = simulation.setup_reset(m, rst, m.make_reset(), period=100)

    init.add(
        Delay(1000 * 100),
        Systask('finish'),
    )

    return m


if __name__ == '__main__':
    test = mkTest()
    verilog = test.to_verilog('tmp.v')
    print(verilog)

    sim = simulation.Simulator(test)
    rslt = sim.run()
    print(rslt)

    # sim.view_waveform()
